from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from .data import _get_connection
from .plotting import _init_plot, _draw_plot, colormap
from .isotope import Isotope
from .library import Library

class Reaction(object):
	"""Cross section data for nuclear reactions

	Contains reaction cross sections as a function of incident energy,
	and some useful methods for manipulating cross section data, such as
	flux-averages, integrated cross-sections, and interpolation.  All 
	cross sections (and uncertainties) are in mb, and all energies are
	in MeV.
	
	Parameters
	----------
	reaction_name : str
		Name of the reaction, in nuclear reaction notation. E.g. '115IN(n,g)',
		'235U(n,f)', '139LA(p,x)134CE', 'Ra-226(n,2n)Ra-225', 'Al-27(n,a)', etc.

	library : str, optional
		Name of the library to use, or 'best' (default).

	Attributes
	----------
	target : str
		The target nucleus.  Some libraries support natural elements, e.g. 'natEl'. 

	incident : str
		Incident particle. E.g. 'n', 'p', 'd'. 

	outgoing : str
		Outgoing particle, or reaction shorthand.  E.g. '2n', 'd', 'f', 'inl', 'x'.
		Will always be 'x' for (TENDL) residual product libraries. 

	product : str
		The product isotope. 

	eng : np.ndarray
		Incident particle energy, in MeV. 

	xs : np.ndarray
		Reaction cross section, in mb. 

	unc_xs : np.ndarray
		Uncertainty in the cross section, in mb.  If not provided by the
		library, default is zeros of same shape as xs. 

	name : str
		Name of the reaction in nuclear reaction notation. 

	library : ci.Library
		Nuclear reaction library.  printing `rx.library.name` will give the
		name of the library. 

	TeX : str
		LaTeX formatted reaction name. 


	Examples
	--------
	>>> rx = ci.Reaction('226RA(n,2n)')
	>>> print(rx.library.name)
	ENDF/B-VII.1
	>>> rx = ci.Reaction('226RA(n,x)225RA')
	>>> print(rx.library.name)
	TENDL-2015
	>>> rx = ci.Reaction('115IN(n,inl)')
	>>> print(rx.library.name)
	IRDFF-II

	"""

	def __init__(self, reaction_name, library='best'):
		self.target, p = tuple(reaction_name.split('('))
		p, self.product = tuple(p.split(')'))
		self.incident, self.outgoing = tuple(p.split(','))
		self.incident, self.outgoing = self.incident.lower(), self.outgoing.lower()
		self._rx = [self.target, self.incident, self.outgoing, self.product]
		self.name = reaction_name

		if library.lower()=='best':
			if self.incident=='n':
				for lb in ['irdff','endf','iaea','tendl','tendl_n_rp']:
					self.library = Library(lb)
					if lb=='tendl_n_rp':
						self._check(True)
					elif self._check():
						break
			elif self.incident in ['p','d']:
				for lb in ['iaea','tendl_'+self.incident+'_rp']:
					self.library = Library(lb)
					if lb=='tendl_d_rp':
						self._check(True)
					elif self._check():
						break
			else:
				self.library = Library('iaea')
				self._check(True)
		else:
			self.library = Library(library)
			self._check(True)

		self.name = self.library.search(*self._rx)[0]
		q = self.library.retrieve(*self._rx)
		self.eng = q[:,0]
		self.xs = q[:,1]
		if q.shape[1]==3:
			self.unc_xs = q[:,2]
		else:
			self.unc_xs = np.zeros(len(self.xs))
		self._interp = None
		self._interp_unc = None

		try:
			if 'nat' not in self.target:
				tg = Isotope(self.target).TeX
			else:
				tg = r'$^{nat}$'+self.target[3:].title()
			prd = Isotope(self.product).TeX if self.product else ''

			self.TeX = '{0}({1},{2}){3}'.format(tg, self.incident, self.outgoing, prd)

		except:
			self.TeX = reaction_name

	def _check(self, err=False):
		c = len(self.library.search(*self._rx))==1
		if err and not c:
			raise ValueError('Reaction '+self.name+' not found or not unique.')
		return c

	def __str__(self):
		return self.name
		
	def interpolate(self, energy):
		"""Interpolated cross section

		Linear interpolation of the reaction cross section along the
		input energy grid.

		Parameters
		----------
		energy : array_like
			Incident particle energy, in MeV.

		Returns
		-------
		cross_section : np.ndarray
			Interpolated cross section, in mb.

		Examples
		--------
		>>> rx = ci.Reaction('115IN(n,g)', 'IRDFF')
		>>> print(rx.interpolate(0.5))
		161.41656650941306
		>>> print(rx.interpolate([0.5, 1.0, 5.0]))
		[161.41646651 171.81486757 8.8822]

		""" 

		if self._interp is None:
			kind = 'linear'
			fv = 'extrapolate'
			i = 0
			if self.library.name.lower().startswith('tendl'):
				kind = 'quadratic'
				fv = 0.0
				ix = np.where(self.xs>0)[0]
				if len(ix)>0:
					i = max((ix[0]-1, 0))
					if len(self.xs)-i<5:
						kind = 'linear'
			self._interp = interp1d(self.eng[i:], self.xs[i:], bounds_error=False, fill_value=fv, kind=kind)
		_interp = self._interp(energy)
		return np.where(_interp>0, _interp, 0.0)

	def interpolate_unc(self, energy):
		"""Uncertainty in interpolated cross section

		Linear interpolation of the uncertainty in the reaction cross section
		along the input energy grid, for libraries where uncertainties are provided.

		Parameters
		----------
		energy : array_like
			Incident particle energy, in MeV.

		Returns
		-------
		unc_cross_section : np.ndarray
			Uncertainty in the interpolated cross section, in mb.

		Examples
		--------
		>>> rx = ci.Reaction('115IN(n,g)', 'IRDFF')
		>>> print(rx.interpolate_unc(0.5))
		3.9542683715745546
		>>> print(rx.interpolate_unc([0.5, 1.0, 5.0]))
		[3.95426837 5.88023936 0.4654]

		""" 

		if self._interp_unc is None:
			self._interp_unc = interp1d(self.eng, self.unc_xs, bounds_error=False, fill_value=0.0)
		return self._interp_unc(energy)
		
	def integrate(self, energy, flux, unc=False):
		"""Reaction flux integral

		Integrate the product of the cross section and flux along the input energy grid.

		Parameters
		----------
		energy : array_like
			Incident particle energy, in MeV.

		flux : array_like
			Incident particle flux as a function of the input energy grid.

		unc : bool, optional
			If `True`, returns the both the flux integral and the uncertainty. If `False`,
			just the flux integral is returned. Default `False`.

		Returns
		-------
		xs_integral : float or tuple
			Reaction flux integral if `unc=False` (default), or reaction flux integral
			and uncertainty, if `unc=True`.

		Examples
		--------
		>>> x = ci.Reaction('Ni-58(n,p)')
		>>> eng = np.linspace(1, 5, 20)
		>>> phi = np.ones(20)
		>>> print(rx.integrate(eng, phi))
		833.4435915974148
		>>> print(rx.integrate(eng, phi, unc=True))
		(833.4435915974148, 19.91851943674977)

		"""

		E = np.asarray(energy)
		phisig = np.asarray(flux)*self.interpolate(E)
		if unc:
			unc_phisig = np.asarray(flux)*self.interpolate_unc(E)
			return np.sum(0.5*(E[1:]-E[:-1])*(phisig[:-1]+phisig[1:])), np.sum(0.5*(E[1:]-E[:-1])*(unc_phisig[:-1]+unc_phisig[1:]))
		return np.sum(0.5*(E[1:]-E[:-1])*(phisig[:-1]+phisig[1:]))
		
	def average(self, energy, flux, unc=False):
		"""Flux averaged reaction cross section

		Calculates the flux-weighted average reaction cross section, using the
		input flux and energy grid.

		Parameters
		----------
		energy : array_like
			Incident particle energy, in MeV.

		flux : array_like
			Incident particle flux as a function of the input energy grid.

		unc : bool, optional
			If `True`, returns the both the flux average cross section and the uncertainty. If `False`,
			just the average cross section is returned. Default `False`.

		Returns
		-------
		average_xs : float or tuple
			Flux-averaged reaction cross section if `unc=False` (default), or average
			and uncertainty, if `unc=True`.

		Examples
		--------
		>>> rx = ci.Reaction('Ni-58(n,p)')
		>>> eng = np.linspace(1, 5, 20)
		>>> phi = np.ones(20)
		>>> print(rx.average(eng, phi))
		208.3608978993537
		>>> print(rx.average(eng, phi, unc=True))
		(208.3608978993537, 4.979629859187442)

		"""

		E, phi = np.asarray(energy), np.asarray(flux)
		phisig = phi*self.interpolate(E)
		dE = E[1:]-E[:-1]
		if unc:
			unc_phisig = np.asarray(flux)*self.interpolate_unc(E)
			return np.sum(0.5*dE*(phisig[:-1]+phisig[1:]))/np.sum(0.5*dE*(phi[:-1]+phi[1:])), np.sum(0.5*dE*(unc_phisig[:-1]+unc_phisig[1:]))/np.sum(0.5*dE*(phi[:-1]+phi[1:]))
		return np.sum(0.5*dE*(phisig[:-1]+phisig[1:]))/np.sum(0.5*dE*(phi[:-1]+phi[1:]))
		
	def plot(self, energy=None, label='reaction', title=False, **kwargs):
		"""Plot the cross section

		Plots the energy differential cross section.

		Parameters
		----------
		energy : array_like, optional
			Energy grid along which to plot the cross section.  If None, the
			energy grid provided by the library will be used. 

		label : str, optional
			Axes label.  If label='reaction', the label will be the reaction name.
			If 'library', it will be the name of the cross section library.
			If 'both', then the reaction name and library will be given.  If
			none of these options, pyplot will be called with `ax.plot(..., label=label)`.

		title : bool, optional
			Display the reaction name as the plot title.  Default, False.

		Other Parameters
		----------------
		**kwargs
			Optional keyword arguments for plotting.  See the 
			plotting section of the curie API for a complete
			list of kwargs.

		Examples
		--------
		>>> rx = ci.Reaction('115IN(n,g)')
		>>> rx.plot(scale='loglog')
		>>> rx = ci.Reaction('35CL(n,p)')
		>>> f, ax = rx.plot(return_plot=True)
		>>> rx = ci.Reaction('35CL(n,el)')
		>>> rx.plot(f=f, ax=ax, scale='loglog')

		"""

		f, ax = _init_plot(**kwargs)

		if title:
			ax.set_title(self.TeX)

		if label is not None:
			if label.lower() in ['both','library','reaction']:
				label = {'both':'{0}\n({1})'.format(self.TeX, self.library.name),'library':self.library.name,'reaction':self.TeX}[label.lower()]

		unc_xs = None
		if energy is None:
			if self.library.name.lower().startswith('tendl'):
				eng = np.linspace(min(self.eng), max(self.eng), 801)
				xs = self.interpolate(eng)
			else:
				eng, xs = self.eng, self.xs
				if np.any(self.unc_xs>0):
					unc_xs = self.unc_xs
		else:
			eng, xs = np.asarray(energy), self.interpolate(energy)
			ux = self.interpolate_unc(energy)
			if np.any(ux>0):
				unc_xs = ux

		line, = ax.plot(eng, xs, label=label)
		if unc_xs is not None:
			ax.fill_between(eng, xs+unc_xs, xs-unc_xs, facecolor=line.get_color(), alpha=0.5)

		if self.library.name.lower().startswith('tendl'):
			wh = np.where((self.eng>=min(eng))&(self.eng<=max(eng)))
			elib = self.eng[wh]
			xslib = self.xs[wh]
			ax.plot(elib, xslib, ls='None', marker='o', color=line.get_color())

		ax.set_xlabel('Incident Energy (MeV)')
		ax.set_ylabel('Cross Section (mb)')

		if label:
			ax.legend(loc=0)

		return _draw_plot(f, ax, **kwargs)
		


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re
import numpy as np
import pandas as pd

from .data import _get_connection

ELEMENTS = ['n','H','He','Li','Be','B','C','N','O','F','Ne','Na','Mg','Al',
			'Si','P','S','Cl','Ar','K','Ca','Sc','Ti','V','Cr','Mn','Fe','Co',
			'Ni','Cu','Zn','Ga','Ge','As','Se','Br','Kr','Rb','Sr','Y','Zr',
			'Nb','Mo','Tc','Ru','Rh','Pd','Ag','Cd','In','Sn','Sb','Te','I',
			'Xe','Cs','Ba','La','Ce','Pr','Nd','Pm','Sm','Eu','Gd','Tb','Dy',
			'Ho','Er','Tm','Yb','Lu','Hf','Ta','W','Re','Os','Ir','Pt','Au',
			'Hg','Tl','Pb','Bi','Po','At','Rn','Fr','Ra','Ac','Th','Pa','U',
			'Np','Pu','Am','Cm','Bk','Cf','Es','Fm','Md','No','Lr','Rf','Db',
			'Sg','Bh','Hs','Mt','Ds','Rg','Cn','Nh','Fl','Mc','Lv','Ts']

class Isotope(object):
	"""Retrieve isotopic structure and decay data

	The Isotope class provides isotope specific nuclear data, such as masses,
	abundances, half-lives, decay data and fission yields.  The main data souces 
	are NuDat 2.0, ENDF/B-VII.0 and the nuclear wallet cards.  Where conflicts
	were found preference was given to NuDat, then ENDF, then wallet cards.
	
	Parameters
	----------
	isotope : str
		Name of the isotope/isomer.  For element El of mass AAA, and (optionally)
		isomeric state m# (where # is an integer starting with 1 for the first isomer)
		or g for ground state, the name must be formatted as either AAAELm# or El-AAAm#
		or EL-AAAm#. If the isomeric state isn't specified, the ground state is assumed,
		and if just m is given, the first isomer is assumed.

	Attributes
	----------
	element : str
		Elemental symbol, e.g. Xe, Ar, Kr

	A : int
		Mass number A of isotope

	isomer : str
		Isomeric state, e.g. g, m1, m2

	name : str
		Formatted isotope name

	E_level : float
		Energy level of the state, in MeV

	J_pi : str
		Spin and partiy of the state

	Z : int
		Atomic number Z of the isotope

	N : int
		Neutron number N of the isotope

	dc : float
		Isotope decay constant in units of 1/seconds

	stable : bool
		`True` if isotope is stable

	mass : float
		Atomic mass of the isotope in amu

	abundance : float
		Natural abundance in percent.

	unc_abundance : float
		Uncertainty in natural abundance in percent.

	Delta : float
		Mass excess of the isotope in MeV

	decay_products : dict
		Dictionary of decay products and their absolute branching ratios as `{'istp':BR}`

	TeX : str
		LaTeX formatted isotope name

	Examples
	--------
	>>> ip = ci.Isotope('115INm')
	>>> ip = ci.Isotope('Co-60')
	>>> ip = ci.Isotope('58NI')
	>>> print(ip.abundance)
	68.077
	
	>>> ip = ci.Isotope('135CEm')
	>>> print(ip.dc)
	0.03465735902799726

	>>> ip = ci.Isotope('235U')
	>>> print(ip.half_life('My'))
	703.798767967


	"""

	def __init__(self, isotope):
		self._parse_itp_name(isotope)

		df = pd.read_sql('SELECT * FROM chart WHERE name="{}"'.format(self.name), _get_connection('decay'))
		self.E_level = float(df['E_level'][0])
		self.J_pi = str(df['J_pi'][0])
		self.stable = bool(df['stable'][0])
		if self.stable:
			self.dc = 0.0
			self.decay_products = {}
		self.mass = float(df['amu'][0])
		if df['abundance'][0] is not None:
			self.abundance = float(df['abundance'][0])
		else:
			self.abundance = 0.0
		if df['Delta'][0] is not None:
			self.Delta = float(df['Delta'][0])
		else:
			self.Delta = None

		if self.abundance>0.0:
			self.unc_abundance = float(df['unc_abundance'][0])
		else:
			self.unc_abundance = 0.0


		self._SFY = None

		if not self.stable:
			self._t_half = float(df['half_life'][0])
			self._unc_t_half = float(df['unc_half_life'][0])
			self.dc = np.log(2.0)/self._t_half

			self.decay_products = {}
			for mode in str(df['decay_mode'][0]).split(','):
				if len(mode.split(':'))!=3:
					continue
				dcy, itp, br = tuple(mode.split(':'))
				br = float(br)
				if itp=='SFY':
					for n,y in self.get_SFY().iterrows():
						self.decay_products[str(y['daughter'])] = y['yield']*br
				else:
					self.decay_products[str(itp)] = br


	def _parse_itp_name(self, istp):
		if istp=='1n' or istp=='1ng' or istp=='n':
			self.element, self.A, self.isomer = 'n', 1, 'g'
			self.name = '1ng'
			self._short_name = '1n'
			
		elif '-' in istp:
			el, A = tuple(istp.split('-'))
			if 'm' in A:
				A, m = tuple(A.split('m'))
				self.isomer = 'm'+m
				if self.isomer=='m':
					self.isomer = 'm1'

			elif 'g' in A:
				A = A[:-1]
				self.isomer = 'g'
			else:
				self.isomer = 'g'

			self.A = int(A)
			self.element = el.title()

		else:
			self.element = ''.join(re.findall('[A-Z]+', istp)).title()
			if istp.startswith('nat'):
				self.A = 'nat'
			else:
				self.A = int(istp.split(self.element.upper())[0])

			self.isomer = istp.split(self.element.upper())[1]
		
		if self.isomer=='' and type(self.A)==int:
			self.isomer = 'g'
		if self.isomer=='m':
			self.isomer = 'm1'

		if self.element!='n':
			self.name = str(self.A)+self.element.upper()+self.isomer
			self._short_name = str(self.A)+self.element.upper()

		self.Z = ELEMENTS.index(self.element)
		if type(self.A)==int:
			self.N = self.A-self.Z

		self.TeX = r'$^{'+str(self.A)+(self.isomer if self.isomer!='g' else '')+r'}$'+self.element


	def __str__(self):
		return self.name

	def half_life(self, units='s', unc=False):
		"""Half-life of isotope

		Returns isotope half-life, in specified units, with or without uncertainty.

		Parameters
		----------
		units : str, optional
			Units for which to return half-life. Options are 
			ns, us, ms, s, m, h, d, y, ky, My, Gy

		unc : bool, optional
			If `True`, uncertainty in half life (in specified units) will also be returned

		Returns
		-------
		half_life : float or 2-tuple of floats
			Half-life in specified units, with uncertainty only if `unc=True`

		Examples
		--------
		>>> ip = ci.Isotope('67CU')
		>>> print(ip.half_life('d'))
		2.57625
		>>> print(ip.half_life('h', True))
		(61.83, 0.12)

		"""

		if units in ['hr','min','yr','sec']:
			units = {'hr':'h','min':'m','yr':'y','sec':'s'}[units]

		half_conv = {'ns':1E-9, 'us':1E-6, 'ms':1E-3,
					's':1.0, 'm':60.0, 'h':3600.0,
					'd':86400.0, 'y':31557.6E3, 'ky':31557.6E6,
					'My':31557.6E9, 'Gy':31557.6E12}[units]

		if self.stable:
			return (np.inf,0.0) if unc else np.inf

		if unc:
			return self._t_half/half_conv, self._unc_t_half/half_conv

		return self._t_half/half_conv

	def decay_const(self, units='s', unc=False):
		"""Decay constant of isotope

		Returns isotope decay constant, in specified units, with or without uncertainty

		Parameters
		----------
		units : str, optional
			Units for which to return decay constant. Options are 
			ns, us, ms, s, m, h, d, y, ky, My, Gy
			Actual value will be given in units of inverse time

		unc : str, optional
			If `True`, uncertainty in decay constant (in specified units) will also be returned

		Returns
		-------
		decay_const : float or 2-tuple of floats
			Decay constant in specified units, with uncertainty only if `unc=True`

		Examples
		--------
		>>> ip = ci.Isotope('221AT')
		>>> print(ip.decay_const())
		0.0050228056562314875
		>>> print(ip.decay_const('m', True))
		(0.3013683393738893, 0.026205942554251245)

		"""

		if self.stable:
			return (0.0, 0.0) if unc else 0.0

		if unc:
			t_half, unc_t_half = self.half_life(units, True)
			dc = np.log(2.0)/t_half
			return dc, dc*(unc_t_half/t_half)

		return np.log(2.0)/self.half_life(units)

	def optimum_units(self):
		"""Appropriate units for half-life 

		Returns
		-------
		units : str
			Units that will represent half-life as a number greater than 1, but with as few digits as possible.

		Examples
		--------
		>>> ip = ci.Isotope('226RA')
		>>> print(ip.half_life())
		50492200000.0
		>>> print(ip.optimum_units())
		y
		>>> print(ip.half_life(ip.optimum_units()))
		1600.00126752

		"""

		opt = ['ns']
		for units in ['us','ms','s','m','h','d','y']:
			if self.half_life(units)>1.0:
				opt.append(units)
		return opt[-1]

	def get_SFY(self):
		"""Spontaneous fission yields

		Returns the absolute spontaneous fission yields (independent) for a given nuclide,
		where available from ENDF.
		
		Returns
		-------
		sfy : pd.DataFrame
			Tabular spontaneous fission yields, with keys 'daughter', 'yield' and 'unc_yield'

		Examples
		--------
		>>> ip = ci.Isotope('238U')
		>>> print(ip.get_SFY())

		"""

		if self._SFY is None:
			self._SFY = pd.read_sql('SELECT daughter, yield, unc_yield FROM SFY WHERE parent="{}"'.format(self.name), _get_connection('decay'))
		return self._SFY

	def get_NFY(self, E):
		"""Neutron induced fission yields

		Returns the absolute neutron induced fission yields (independent) for a given nuclide,
		where available from ENDF.

		Parameters
		----------
		E : float
			Incident neutron energy, in eV. Available energies are 0.0253, 5E5, 2E6 and 14E6.

		Returns
		-------
		nfy : pd.DataFrame
			Tabular neutron induced fission yields, with keys 'daughter', 'yield' and 'unc_yield'

		Examples
		--------
		>>> ip = ci.Isotope('235U')
		>>> print(ip.get_NFY(E=0.0253))

		"""

		return pd.read_sql('SELECT daughter, yield, unc_yield FROM NFY WHERE parent="{0}" AND energy={1}'.format(self.name, E), _get_connection('decay'))


	def gammas(self, I_lim=None, E_lim=None, xrays=False, dE_511=0.0, istp_col=False):
		"""Gamma-rays emitted by the decay of the isotope

		Returns a DataFrame of gamma-ray energies, intensities and intensity-uncertainties
		based on an optional set of selection critieria.

		Parameters
		----------
		I_lim : float or 2-tuple of floats, optional
			Limits on the intensity of the gamma-rays, in percent.  If a single float
			is given, this is a lower bound, if a 2-tuple then (upper, lower) bounds.

		E_lim : float or 2-tuple of floats, optional
			Limits on the energy of the gamma-rays, in keV.  If a single float is given,
			this is a lower bound, if a 2-tuple then (upper, lower) bounds.

		xrays : bool, optional
			If `True`, retrieved data will include x-rays.  Default `False`.

		dE_511 : float, optional
			Filter out gamma-rays that are within `dE_511` keV of the annihilation peak.
			Default 0.0.

		Returns
		-------
		gammas : pd.DataFrame
			Tabular gamma-ray data, with keys 'energy', 'intensity' and 'unc_intensity'. Units
			of energy are in keV and units of intensity are in percent.

		Examples
		--------
		>>> ip = ci.Isotope('Co-60')
		>>> print(ip.gammas(I_lim=1.0))
		     energy  intensity  unc_intensity
		0  1173.228    99.8500         0.0300
		1  1332.492    99.9826         0.0006

		>>> ip = ci.Isotope('64CU')
		>>> print(ip.gammas())
			energy  intensity  unc_intensity
		0   511.00     35.200          0.400
		1  1345.77      0.475          0.011
		>>> print(ip.gammas(xrays=True, dE_511=1.0))
			 energy  intensity  unc_intensity
		0     0.850      0.489          0.024
		1     7.461      4.740          0.240
		2     7.478      9.300          0.400
		3     8.265      1.120          0.050
		4     8.265      0.580          0.030
		5  1345.770      0.475          0.011

		"""

		ip, im = str(self.A)+self.element.upper(), self.isomer

		if xrays:
			df = pd.read_sql('SELECT energy, intensity, unc_intensity FROM gammas WHERE isotope="{0}" AND isomer="{1}"'.format(ip, im), _get_connection('decay'))
		else:
			df = pd.read_sql('SELECT energy, intensity, unc_intensity FROM gammas WHERE isotope="{0}" AND isomer="{1}" AND notes NOT LIKE "%XR%"'.format(ip, im), _get_connection('decay'))

		df = df[np.abs(df['energy']-511.0)>=dE_511]

		if I_lim is not None:
			if type(I_lim)==float or type(I_lim)==int:
				df = df[df['intensity']>=I_lim]
			else:
				df = df[(df['intensity']>=I_lim[0])&(df['intensity']<=I_lim[1])]

		if E_lim is not None:
			if type(E_lim)==float or type(E_lim)==int:
				df = df[df['energy']>=E_lim]
			else:
				df = df[(df['energy']>=E_lim[0])&(df['energy']<=E_lim[1])]

		if istp_col:
			df['isotope'] = self.name

		return df.sort_values(by=['energy']).reset_index(drop=True)

	def electrons(self, I_lim=None, E_lim=None, CE_only=False, Auger_only=False):
		"""Electrons (conversion and Auger) emitted by the decay of the isotope

		Returns a DataFrame of electron energies, intensities and intensity-uncertainties
		based on an optional set of selection critieria.

		Parameters
		----------
		I_lim : float or 2-tuple of floats, optional
			Limits on the intensity of the electrons, in percent.  If a single float
			is given, this is a lower bound, if a 2-tuple then (upper, lower) bounds.

		E_lim : float or 2-tuple of floats, optional
			Limits on the energies of the electrons, in keV.  If a single float
			is given, this is a lower bound, if a 2-tuple then (upper, lower) bounds.

		CE_only : bool
			If `True`, only conversion electrons will be returned. Default `False`

		Auger_only : bool
			If `True`, only Auger electrons will be returned. Default `False`.

		Returns
		-------
		electrons : pd.DataFrame
			Tabular electron data, with keywords 'energy', 'intensity', 'unc_intensity'. Units
			of energy are in keV and units of intensity are in percent.

		Examples
		--------
		>>> ip = ci.Isotope('Pt-193m')
		>>> print(ip.electrons(I_lim=5.0, E_lim=(10.0, 130.0)))
		    energy  intensity  unc_intensity
		0   11.912       17.1          0.855
		1   57.110       15.5          0.775
		2  121.620       60.0          3.000

		"""

		ip, im = str(self.A)+self.element.upper(), self.isomer

		if CE_only:
			df = pd.read_sql('SELECT energy, intensity, unc_intensity FROM electrons WHERE isotope="{0}" AND isomer="{1}" AND notes LIKE "%CE%"'.format(ip, im), _get_connection('decay'))
		elif Auger_only:
			df = pd.read_sql('SELECT energy, intensity, unc_intensity FROM electrons WHERE isotope="{0}" AND isomer="{1}" AND notes LIKE "%Auger%"'.format(ip, im), _get_connection('decay'))
		else:
			df = pd.read_sql('SELECT energy, intensity, unc_intensity FROM electrons WHERE isotope="{0}" AND isomer="{1}"'.format(ip, im), _get_connection('decay'))

		if I_lim is not None:
			if type(I_lim)==float or type(I_lim)==int:
				df = df[df['intensity']>=I_lim]
			else:
				df = df[(df['intensity']>=I_lim[0])&(df['intensity']<=I_lim[1])]

		if E_lim is not None:
			if type(E_lim)==float or type(E_lim)==int:
				df = df[df['energy']>=E_lim]
			else:
				df = df[(df['energy']>=E_lim[0])&(df['energy']<=E_lim[1])]

		return df.sort_values(by=['energy']).reset_index(drop=True)

	def beta_minus(self, I_lim=None, Endpoint_lim=None):
		"""Electrons from beta-minus decay

		Returns a DataFrame of beta-minus mean and enpoint energies, intensities and
		intensity-uncertainties based on an optional set of selection critieria.

		Parameters
		----------
		I_lim : float or 2-tuple of floats, optional
			Limits on the intensity of the beta-minus decay, in percent.  If a single float
			is given, this is a lower bound, if a 2-tuple then (upper, lower) bounds.

		Endpoint_lim : float or 2-tuple of floats, optional
			Limits on the endpoint energy of the beta-minus decay, in keV.  If a single float
			is given, this is a lower bound, if a 2-tuple then (upper, lower) bounds.

		Returns
		-------
		beta_minus : pd.DataFrame
			Tabular beta-minus data, with keywords 'mean_energy', 'endpoint_energy', 'intensity',
			'unc_intensity'. Units of energy are in keV and units of intensity are in percent.

		Examples
		--------
		>>> ip = ci.Isotope('35S')
		>>> print(ip.beta_minus())
		   mean_energy  endpoint_energy  intensity  unc_intensity
		0       48.758           167.33      100.0            5.0

		"""

		ip, im = str(self.A)+self.element.upper(), self.isomer

		df = pd.read_sql('SELECT mean_energy, endpoint_energy, intensity, unc_intensity FROM beta_minus WHERE isotope="{0}" AND isomer="{1}"'.format(ip, im), _get_connection('decay'))

		if I_lim is not None:
			if type(I_lim)==float or type(I_lim)==int:
				df = df[df['intensity']>=I_lim]
			else:
				df = df[(df['intensity']>=I_lim[0])&(df['intensity']<=I_lim[1])]

		if Endpoint_lim is not None:
			if type(Endpoint_lim)==float or type(Endpoint_lim)==int:
				df = df[df['endpoint_energy']>=Endpoint_lim]
			else:
				df = df[(df['endpoint_energy']>=Endpoint_lim[0])&(df['endpoint_energy']<=Endpoint_lim[1])]

		return df.sort_values(by=['endpoint_energy']).reset_index(drop=True)

	def beta_plus(self, I_lim=None, Endpoint_lim=None):
		"""Positrons from beta-plus decay

		Returns a DataFrame of beta-plus mean and enpoint energies, intensities and
		intensity-uncertainties based on an optional set of selection critieria.

		Parameters
		----------
		I_lim : float or 2-tuple of floats, optional
			Limits on the intensity of the beta-plus decay, in percent.  If a single float
			is given, this is a lower bound, if a 2-tuple then (upper, lower) bounds.

		Endpoint_lim : float or 2-tuple of floats, optional
			Limits on the endpoint energy of the beta-plus decay, in keV.  If a single float
			is given, this is a lower bound, if a 2-tuple then (upper, lower) bounds.

		Returns
		-------
		beta_plus : pd.DataFrame
			Tabular beta-plus data, with keywords 'mean_energy', 'endpoint_energy', 'intensity',
			'unc_intensity'. Units of energy are in keV and units of intensity are in percent.

		Examples
		--------
		>>> ip = ci.Isotope('18F')
		>>> print(ip.beta_plus())
		   mean_energy  endpoint_energy  intensity  unc_intensity
		0        249.8            633.5      96.73           0.04

		"""

		ip, im = str(self.A)+self.element.upper(), self.isomer

		df = pd.read_sql('SELECT mean_energy, endpoint_energy, intensity, unc_intensity FROM beta_plus WHERE isotope="{0}" AND isomer="{1}"'.format(ip, im), _get_connection('decay'))

		if I_lim is not None:
			if type(I_lim)==float or type(I_lim)==int:
				df = df[df['intensity']>=I_lim]
			else:
				df = df[(df['intensity']>=I_lim[0])&(df['intensity']<=I_lim[1])]

		if Endpoint_lim is not None:
			if type(Endpoint_lim)==float or type(Endpoint_lim)==int:
				df = df[df['endpoint_energy']>=Endpoint_lim]
			else:
				df = df[(df['endpoint_energy']>=Endpoint_lim[0])&(df['endpoint_energy']<=Endpoint_lim[1])]

		return df.sort_values(by=['endpoint_energy']).reset_index(drop=True)

	def alphas(self, I_lim=None, E_lim=None):
		"""Alpha-particles emitted by the decay of the isotope

		Returns a DataFrame of alpha-particle energies, intensities and intensity-uncertainties
		based on an optional set of selection critieria.

		Parameters
		----------
		I_lim : float or 2-tuple of floats, optional
			Limits on the intensity of the alphas, in percent.  If a single float
			is given, this is a lower bound, if a 2-tuple then (upper, lower) bounds.

		E_lim : float or 2-tuple of floats, optional
			Limits on the energy of the alphas, in keV.  If a single float is given,
			this is a lower bound, if a 2-tuple then (upper, lower) bounds.

		Returns
		-------
		alphas : pd.DataFrame
			Tabular alpha emission data, with keys 'energy', 'intensity' and 'unc_intensity'. Units
			of energy are in keV and units of intensity are in percent.

		Examples
		--------
		>>> ip = ci.Isotope('210PO')
		>>> print(ip.alphas(I_lim=1.0))
		    energy  intensity  unc_intensity
		0  5304.33      100.0            5.0

		"""

		ip, im = str(self.A)+self.element.upper(), self.isomer

		df = pd.read_sql('SELECT energy, intensity, unc_intensity FROM alphas WHERE isotope="{0}" AND isomer="{1}"'.format(ip, im), _get_connection('decay'))

		if I_lim is not None:
			if type(I_lim)==float or type(I_lim)==int:
				df = df[df['intensity']>=I_lim]
			else:
				df = df[(df['intensity']>=I_lim[0])&(df['intensity']<=I_lim[1])]

		if E_lim is not None:
			if type(E_lim)==float or type(E_lim)==int:
				df = df[df['energy']>=E_lim]
			else:
				df = df[(df['energy']>=E_lim[0])&(df['energy']<=E_lim[1])]

		return df.sort_values(by=['energy']).reset_index(drop=True)

	def dose_rate(self, activity=1.0, distance=30.0, units='R/hr'):
		"""Dose-rate from a point source of the isotope

		Assuming a point source with no attenuation, this function calculates the dose-rate
		from each emitted particle type using a specified source activity and distance.

		Parameters
		----------
		activity : float, optional
			Source activity in Bq. Default is 1.0

		distance : float, optional
			Source distance in cm. Default is 30.0

		units : str, optional
			Desired units of dose rate. Format: `dose/time` where dose is one of R (Roentgen), Sv, Gy, with
			the Si prefixes u, m, k, and M supported, and time is in the same format described under
			the `Isotope.half_life()` function. E.g. mR/hr, uSv/m, kGy/y. Default is 'R/hr'

		Returns
		-------
		dose_rate : dict
			Dictionary of dose from each of the following particle type: 'gamma', 'electron', 
			'beta_minus', 'beta_plus', 'alpha', and 'total'.  Units are specified as an input.

		Examples
		--------
		>>> ip = ci.Isotope('Co-60')
		>>> print(ip.dose_rate(activity=3.7E10, units='R/hr')) # 1 Ci of Co-60 at 30 cm
		{'beta_plus': 0.0, 'alphas': 0.0, 'gammas': 52035.28424827692, 
		'electrons': 65.66251805557148, 'beta_minus': 5536.541902410433, 'total': 57637.488668742924}

		"""

		def beta2(E_MeV, m_amu):
			return 1.0-(1.0/(1.0+(E_MeV/m_amu))**2)

		def e_range(E_keV):
			dEdx = lambda b2, t: 0.17*((np.log(3.61E5*t*np.sqrt(t+2)))+(0.5*(1-b2)*(1+t**2/8-(2*t+1)*np.log(2)))-4.312)/b2
			E = np.linspace(1E-12, E_keV*1E-3, 100)
			return np.trapz(1.0/((1.0+(E*7.22/800.0))*dEdx(beta2(E, 0.5109), E/0.5109)), E)

		def pos_range(E_keV):
			dEdx = lambda b2, t: 0.17*((np.log(3.61E5*t*np.sqrt(t+2)))+(np.log(2)-(b2/24)*(23+14/(t+2)+10/(t+2)**2+4/(t+2)**3))-4.312)/b2
			E = np.linspace(1E-12, E_keV*1E-3, 100)
			return np.trapz(1.0/((1.0+(E*7.22/800.0))*dEdx(beta2(E, 0.5109), E/0.5109)), E)

		def alpha_range(E_keV):
			dEdx = lambda b2: 0.17*4.0*((np.log(1.02E6*b2/(1.0-b2))-b2)-4.312)/b2
			return np.trapz(1.0/dEdx(beta2(np.linspace(1E-9, E_keV*1E-3, 100), 3.7284E3)), np.linspace(1E-9, E_keV*1E-3, 100))

		dose = {}

		gm = self.gammas(xrays=True, dE_511=0.0)
		al = self.alphas()
		bm = self.beta_minus()
		bp = self.beta_plus()
		el = self.electrons()

		dose['gammas'] = 3.9005E-16*np.sum(gm['energy']*gm['intensity'])*activity/distance**2

		dose['alphas'] = 1.45367E-14*np.sum([al['intensity'][n]*e/alpha_range(e) for n,e in enumerate(al['energy'].to_numpy())])*activity/distance**2

		dose['beta_minus'] = 1.45367E-14*np.sum([bm['intensity'][n]*e/e_range(e) for n,e in enumerate(bm['mean_energy'].to_numpy())])*activity/distance**2
		dose['beta_plus'] = 1.45367E-14*np.sum([bp['intensity'][n]*e/pos_range(e) for n,e in enumerate(bp['mean_energy'].to_numpy())])*activity/distance**2
		dose['electrons'] = 1.45367E-14*np.sum([el['intensity'][n]*e/e_range(e) for n,e in enumerate(el['energy'].to_numpy())])*activity/distance**2
		

		d_unit,t_unit = tuple(units.split('/'))

		if 'Sv' in d_unit:
			dose['alpha'] = 20.0*dose['alpha']
			d_unit = d_unit.replace('Sv','Gy')

		if t_unit in ['hr','min','yr','sec']:
			t_unit = {'hr':'h','min':'m','yr':'y','sec':'s'}[t_unit]

		t_conv = {'ns':1E-9, 'us':1E-6, 'ms':1E-3,
					's':1.0, 'm':60.0, 'h':3600.0,
					'd':86400.0, 'y':31557.6E3, 'ky':31557.6E6,
					'My':31557.6E9, 'Gy':31557.6E12}[t_unit]

		d_conv = {'R':1.0,'mR':1E3,'kR':1E-3,'uR':1E6,'MR':1E-6,
					'Gy':8.77E-3,'mGy':8.77,'kGy':8.77E-6,'uGy':8.77E3,'MGy':8.77E-9}[d_unit]

		dose = {i:d_conv*t_conv*dose[i] for i in ['gammas','alphas','beta_minus','beta_plus','electrons']}
		dose['total'] = sum([dose[i] for i in ['gammas','alphas','beta_minus','beta_plus','electrons']])

		return dose



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sqlite3

GLOB_CONNECTIONS_DICT = {}

def _data_path(db=''):
	return os.path.realpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', db))

def download(db='all', overwrite=False):
	"""Download nuclear data files as sqlite .db files.

	This function can be used to either download or update the nuclear data
	files required by Curie.  If you wish to update, or if the data files are
	corrupted, run `ci.download(overwrite=True)` to overwrite the existing data.

	Parameters
	----------
	db : str, optional
		Name of database to download, default is 'all'.  Options include:
		'all', 'decay', 'ziegler', 'endf', 'tendl', 'tendl_n_rp', 'tendl_p_rp',
		'tendl_d_rp', 'IRDFF', 'iaea_monitors'.

	overwrite : bool, optional
		If overwrite is `True`, will save write over existing data.  Default is `False`.

	Examples
	--------
	The most common use case will be to (re)download all the data files

	>>> ci.download(overwrite=True)

	Some other use cases:
	To update only the 'endf' library

	>>> ci.download('endf', True)

	Or to download the 'decay' library for the first time

	>>> ci.download('decay')

	"""

	db = db.lower()
	if db in ['all','*']:
		d = ['decay', 'ziegler', 'endf', 'tendl', 'tendl_n_rp', 'tendl_p_rp', 'tendl_d_rp', 'IRDFF', 'iaea_monitors']
	elif db in ['decay']:
		d = ['decay']
	elif db in ['ziegler']:
		d = ['ziegler']
	elif db in ['endf']:
		d = ['endf']
	elif db in ['tendl']:
		d = ['tendl']
	elif db in ['tendl_n_rp', 'tendl_nrp', 'tendl_n', 'nrp', 'rpn']:
		d = ['tendl_n_rp']
	elif db in ['tendl_p_rp', 'tendl_prp', 'tendl_p', 'prp', 'rpp']:
		d = ['tendl_p_rp']
	elif db in ['tendl_d_rp', 'tendl_drp', 'tendl_d', 'drp', 'rpd']:
		d = ['tendl_d_rp']
	elif db in ['irdff']:
		d = ['IRDFF']
	elif db in ['iaea', 'iaea-cpr', 'iaea-monitor', 'cpr', 'iaea_cpr', 'iaea_monitor', 'medical', 'iaea-medical', 'iaea_medical']:
		d = ['iaea_monitors']
	else:
		print('db={} not recognized.'.format(db))
		return

	addr = {'decay':'wwd6b1gk2ge5tgt', 'endf':'tkndjqs036piojm', 'tendl':'zkoi6t2jicc9yqs', 'tendl_d_rp':'x2vfjr7uv7ffex5', 'tendl_n_rp':'n0jjc0dv61j9of9',
				'tendl_p_rp':'ib2a5lrhiwkcro5', 'ziegler':'kq07684wtp890v5', 'iaea_monitors':'lzn8zs6y8zu3v0s', 'IRDFF':'34sgcvt8n57b0aw'}
	
	if not os.path.isdir(_data_path()):
		os.mkdir(_data_path())

	try:
		import urllib2
	except:
		import urllib.request as urllib2

	for i in d:
		fnm = i+'.db'
		if (not os.path.isfile(_data_path(fnm))) or overwrite:
			
			try:
				print('Downloading {}'.format(fnm))
				with open(_data_path(fnm),'wb') as f:
					f.write(urllib2.urlopen('https://www.dropbox.com/s/{0}/{1}?dl=1'.format(addr[i],fnm)).read())
			except Exception as e:
					print(e)
		else:
			print("{0}.db already installed. Run ci.download('{0}', overwrite=True) to overwrite these files.".format(i))



def _get_connection(db='decay'):
	
	def connector(dbnm):
		if os.path.exists(dbnm):
			try:
				if os.path.getsize(dbnm)>0:
					return sqlite3.connect(dbnm)
				else:
					raise ValueError('{} exists but is of zero size.'.format(dbnm))
			except:
				print('Error connecting to {}.'.format(dbnm))
				if _data_path() in dbnm:
					print('Try using ci.download("all", overwrite=True) to update nuclear data files.')
				raise
		else:
			print('WARNING: database {} does not exist, creating new file.'.format(dbnm))
			return sqlite3.connect(dbnm)


	db_nm = db.lower().replace('.db','')
	db_f = None

	if db_nm in ['decay']:
		db_f = 'decay.db'

	elif db_nm in ['ziegler']:
		db_f = 'ziegler.db'

	elif db_nm in ['endf']:
		db_f = 'endf.db'

	elif db_nm in ['tendl']:
		db_f = 'tendl.db'

	elif db_nm in ['tendl_n_rp','tendl_nrp','tendl_n','nrp','rpn']:
		db_f = 'tendl_n_rp.db'

	elif db_nm in ['tendl_p_rp','tendl_prp','tendl_p','prp','rpp']:
		db_f = 'tendl_p_rp.db'

	elif db_nm in ['tendl_d_rp','tendl_drp','tendl_d','drp','rpd']:
		db_f = 'tendl_d_rp.db'

	elif db_nm in ['irdff']:
		db_f = 'IRDFF.db'

	elif db_nm in ['iaea', 'iaea-cpr', 'iaea-monitor', 'cpr', 'iaea_cpr', 'iaea_monitor', 'medical', 'iaea-medical', 'iaea_medical']:
		db_f = 'iaea_monitors.db'


	global GLOB_CONNECTIONS_DICT

	if db_f is not None:
		if db_f not in GLOB_CONNECTIONS_DICT:
			GLOB_CONNECTIONS_DICT[db_f] = connector(_data_path(db_f))
		return GLOB_CONNECTIONS_DICT[db_f]

	else:
		if db not in GLOB_CONNECTIONS_DICT:
			GLOB_CONNECTIONS_DICT[db] = connector(db)
		return GLOB_CONNECTIONS_DICT[db]



def _get_cursor(db='decay'):	
	conn = _get_connection(db)
	return conn.cursor()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import pandas as pd
import datetime as dtm
pd.options.mode.chained_assignment = None

from scipy.optimize import curve_fit

from .data import _get_connection
from .plotting import _init_plot, _draw_plot, colormap
from .isotope import Isotope
from .spectrum import Spectrum

class DecayChain(object):
	"""Radioactive Decay Chain

	Uses the Bateman equations to calculate the activities and number of decays
	from a radioactive decay chain as a function of time, both in production
	and decay.  Also, initial isotope activities and production rates can
	be fit to observed count data, or directly fit to HPGe spectra using the
	`get_counts()` function.
	
	Parameters
	----------
	parent_isotope : str
		Parent isotope in the chain.

	R : array_like, dict, str or pd.DataFrame
		Production rate for each isotope in the decay chain as a function of time.
		If a Nx2 np.ndarray, element n gives the production rate R_n up until
		time t_n for the parent isotope. E.g. If the production rate of the parent
		is 5 for 1 hour, and 8 for 3 hours, the array will be [[5, 1], [8, 4]].  If
		instead time intervals are preferred to a monotonically increasing grid
		of timestamps, set 'timestamp=False'.  In this case the production rate array
		will be [[5, 1], [8, 3]]. (R=5 for 1 hour, R=8 for 3 hours).

		If R is a dict, it specifies the production rate for multiple isotopes, 
		where the keys are the isotopes and the values are type np.ndarray.

		If R is a pd.DataFrame, it must have columns 'R' and 'time', and optionally 'isotope'
		if R>0 for any isotopes other than the parent.  If R is a str, it must be a 
		path to a file where the same data is provided.  Supported file types are
		.csv, .json and .db files, where .json files must be in the 'records' format,
		and .db files must have a table named 'R'.  Also, each isotope must have
		the same time grid, for both timestamp=True and timestamp=False.

	A0 : float or dict
		Initial activity.  If a float, the initial activity of the parent isotope.
		If a dict, the keys are the isotopes for which the values represent the
		initial activity.

	units : str, optional
		Units of time for the chain. Options are 'ns', 'us', 'ms', 's', 'm', 'h', 
		'd', 'y', 'ky', 'My', 'Gy'.  Default is 's'.

	timestamp : bool, optional
		Determines if the 'time' variable in R is to be read as a timestamp-like grid,
		i.e. in monotonically increasing order, or as a series of time intervals.
		Default is `True`.

	Attributes
	----------
	R : pd.DataFrame
		Production rate as a function of time, for each isotope in the chain. This
		will be modified if `fit_R()` is called.

	A0 : dict
		Initial activity of each isotope in the chain.

	isotopes : list
		List of isotopes in the decay chain.

	counts : pd.DataFrame
		Observed counts from isotopes in the decay chain, which can be used
		to determine the initial activities or average production rates using
		the `fit_R()` or `fit_A0()` functions.

	R_avg : pd.DataFrame
		Time-averaged production rate for each isotope where R>0.  This will be
		modified if `fit_R()` is called.

	Examples
	--------
	>>> dc = ci.DecayChain('Ra-225', R=[[1.0, 1.0], [0.5, 1.5], [2.0, 6]], units='d')
	>>> print(dc.isotopes)
	['225RAg', '225ACg', '221FRg', '217ATg', '213BIg', '217RNg', '209TLg', '213POg', '209PBg', '209BIg']
	>>> print(dc.R_avg)
          R_avg isotope
	0  1.708333  225RAg

	>>> dc = ci.DecayChain('152EU', A0=3.7E3, units='h')
	>>> print(ci.isotopes)
	['152EUg', '152GDg', '152SMg']

	"""

	def __init__(self, parent_isotope, R=None, A0=None, units='s', timestamp=True):
		if units.lower() in ['hr','min','yr','sec']:
			units = {'hr':'h','min':'m','yr':'y','sec':'s'}[units.lower()]
		self.units = units
		

		istps = [Isotope(parent_isotope)]
		self.isotopes = [istps[0].name]
		self.R, self.A0, self._counts = None, {self.isotopes[0]:0.0}, None
		self._chain = [[istps[0].decay_const(units), [], []]]
		stable_chain = [False]

		while not all(stable_chain):
			for n in [n for n,ch in enumerate(stable_chain) if not ch]:
				stable_chain[n] = True
				for prod in istps[n].decay_products:
					br = istps[n].decay_products[prod]
					I = Isotope(prod)
					if I.name in self.isotopes:
						self._chain[self.isotopes.index(I.name)][1].append(br)
						self._chain[self.isotopes.index(I.name)][2].append(n)
					else:
						istps.append(I)
						self.isotopes.append(I.name)
						self._chain.append([I.decay_const(units), [br], [n]])
						stable_chain.append(self._chain[-1][0]<1E-12)
						if not stable_chain[-1]:
							self.A0[self.isotopes[-1]] = 0.0
		self._chain = np.array(self._chain, dtype=object)
		self._branches = self._generate_branches()

		if A0 is not None:
			if type(A0)==float or type(A0)==int:
				self.A0[self.isotopes[0]] = float(A0)
			else:
				for i in A0:
					self.A0[self._filter_name(i)] = float(A0[i])


		if R is not None:
			if type(R)==str:
				if R.endswith('.json'):
					self.R = pd.DataFrame(json.loads(open(R).read()))
				elif R.endswith('.csv'):
					self.R = pd.read_csv(R, header=0).fillna(method='ffill')
				elif R.endswith('.db'):
					self.R = pd.read_sql('SELECT * FROM R', _get_connection(R))

				if 'isotope' not in self.R.columns.to_list():
					self.R['isotope'] = self.isotopes[0]

			elif type(R)==pd.DataFrame:
				self.R = R.copy(deep=True)
				if 'isotope' not in self.R.columns.to_list():
					self.R['isotope'] = self.isotopes[0]

			elif type(R)==dict:
				self.R = pd.DataFrame({'isotope':[], 'R':[], 'time':[]})
				for ip in R:
					rate = np.array(R[ip])
					rt = pd.DataFrame({'isotope':self._filter_name(ip),'R':rate[:,0],'time':rate[:,1]})
					self.R = pd.concat([self.R, rt], ignore_index=True).reset_index(drop=True)

			elif type(R)==list or type(R)==np.ndarray:
				R = np.asarray(R)
				self.R = pd.DataFrame({'isotope':self.isotopes[0], 'R':R[:,0], 'time':R[:,1]})

			self.R['isotope'] = [self._filter_name(i) for i in self.R['isotope']]

			if not timestamp:
				for ip in pd.unique(self.R['isotope']):
					self.R.loc[self.R['isotope']==ip,'time'] = np.cumsum(self.R.loc[self.R['isotope']==ip,'time'])

			time = np.insert(np.unique(self.R['time']), 0, [0.0])
			for n,dt in enumerate(time[1:]-time[:-1]):
				_R_dict = {p:self.R[self.R['isotope']==p].iloc[n]['R'] for p in pd.unique(self.R['isotope'])}
				self.A0 = {p:self.activity(p, dt, _R_dict=_R_dict) for p in self.A0}

	def __str__(self):
		return self.isotopes[0]

	def _filter_name(self, istp):
		return Isotope(istp).name

	def _index(self, istp):
		return self.isotopes.index(self._filter_name(istp))

	def _generate_branches(self):
		daughters = {i:[] for i in range(len(self._chain))}
		for i,pars in enumerate(self._chain[1:,2]):
			for p in pars:
				daughters[p].append(i+1)

		branches = [[0]]
		stable_chain = [len(daughters[br[-1]])==0 for br in branches]
		while not all(stable_chain):
			for par in list(set([b[-1] for b in branches])):
				ds = daughters[par]
				if len(ds)==0:
					continue
				if len(ds)>1:
					to_dup = [br for br in branches if br[-1]==par]
					for m in range(len(ds)-1):
						for b in to_dup:
							branches.append(b+[ds[m+1]])
					for br in branches:
						if br[-1]==par:
							br.append(ds[0])
				else:
					for br in branches:
						if br[-1]==par:
							br.append(ds[0])
			stable_chain = [len(daughters[br[-1]])==0 for br in branches]

		br_ratios = []
		for br in branches:
			r = []
			for n,i in enumerate(br[:-1]):
				r.append(self._chain[br[n+1]][1][self._chain[br[n+1]][2].index(i)])
			br_ratios.append(r+[0.0])

		return branches, br_ratios

	def _get_branches(self, istp):
		if self._filter_name(istp) not in self.isotopes:
			return [], []

		m = self._index(istp)
		branches, br_ratios = [], []

		for n,br in enumerate(self._branches[0]):
			if m in br:
				k = br.index(m)
				new_br = np.array(br[:k+1])
				if not any([np.array_equal(b, new_br) for b in branches]):
					branches.append(new_br)
					br_ratios.append(np.array(self._branches[1][n][:k] + [0.0]))
		return br_ratios, branches

	def _r_lm(self, units=None, r_half_conv=False):
		if units is None:
			return 1.0

		if units.lower() in ['hr','min','yr','sec']:
			units = {'hr':'h','min':'m','yr':'y','sec':'s'}[units.lower()]

		half_conv = {'ns':1E-9, 'us':1E-6, 'ms':1E-3,
					's':1.0, 'm':60.0, 'h':3600.0,
					'd':86400.0, 'y':31557.6E3, 'ky':31557.6E6,
					'My':31557.6E9, 'Gy':31557.6E12}

		if r_half_conv:
			return half_conv[units]

		return half_conv[units]/half_conv[self.units]


	def activity(self, isotope, time, units=None, _R_dict=None, _A_dict=None):
		"""Activity of an isotope in the chain

		Computes the activity of a given isotope in the decay chain at a
		given time.  Units of activity are in Bq.  Units of time must be either
		the units for the DecayChain (default 's'), or specified by the `units`
		keyword.

		Parameters
		----------
		isotope : str
			Isotope for which the activity is calculated.

		time : array_like
			Time to calculate the activity.  Units of time must be the same
			as the decay chain, or be given by `units`. Note that if R!=0, time=0 is 
			defined as the end of production time.  Else, if A0!=0, time=0
			is defined as the time at which the specified activities equaled
			A0.  t<0 is not allowed. 

		units : str, optional
			Units of time, if different from the units of the decay chain.

		Returns
		-------
		activity : np.ndarray
			Activity of the given isotope in Bq.

		Examples
		--------
		>>> dc = ci.DecayChain('152EU', A0=3.7E3, units='h')
		>>> print(dc.activity('152EU', time=0))
		3700.0
		>>> print(dc.activity('152EU', time=13.537, units='y'))
		1849.999906346199

		"""

		time = np.asarray(time)
		A = np.zeros(len(time)) if time.shape else np.array(0.0)

		finished = []
		for m,(BR, chain) in enumerate(zip(*self._get_branches(isotope))):
			lm = self._r_lm(units)*self._chain[chain, 0]
			L = len(chain)
			for i in range(L):
				sub = ''.join(map(str, chain[i:]))
				if sub in finished:
					continue
				finished.append(sub)
				# if i==L-1 and m>0: # only add A0 of end isotope once
					# continue

				ip = self.isotopes[chain[i]]
				A0 = self.A0[ip] if _A_dict is None else _A_dict[ip]
				if A0==0.0 and _R_dict is None:
					continue
				A_i = lm[-1]*(A0/lm[i])
				
				B_i = np.prod(lm[i:-1]*BR[i:-1])

				for j in range(i, L):
					K = np.arange(i, L)
					d_lm = lm[K[K!=j]]-lm[j]
					C_j = np.prod(np.where(np.abs(d_lm)>1E-12, d_lm, 1E-12*np.sign(d_lm)))
					A += A_i*B_i*np.exp(-lm[j]*time)/C_j
					if _R_dict is not None:
						if ip in _R_dict:
							if lm[j]>1E-12:
								A += _R_dict[ip]*lm[-1]*B_i*(1.0-np.exp(-lm[j]*time))/(lm[j]*C_j)
							else:
								A += _R_dict[ip]*lm[-1]*B_i*time/C_j
		return A
		
	def decays(self, isotope, t_start, t_stop, units=None, _A_dict=None):
		"""Number of decays in a given time interval

		Computes the number of decays from a given isotope in the
		decay chain in the time interal t_start to t_stop.  The 
		units of t_start and t_stop must be either the same units
		as the decay chain, or be specified by the `units` keyword.

		Parameters
		----------
		isotope : str
			Isotope for which the number of decays is calculated.

		t_start : array_like
			Time of the start of the interval.

		t_stop : array_like
			Time of the end of the interval.

		units : str, optional
			Units of time, if different from the units of the decay chain.

		Returns
		-------
		decays : np.ndarray
			Number of decays

		Examples
		--------
		>>> dc = ci.DecayChain('152EU', A0=3.7E3, units='h')
		>>> print(dc.decays('152EU', t_start=1, t_stop=2))
		13319883.293399204
		>>> print(dc.decays('152EU', t_start=50, t_stop=50.1, units='y'))
		900151618.5228329

		"""

		t_start, t_stop = np.asarray(t_start), np.asarray(t_stop)
		D = np.zeros(len(t_start)) if t_start.shape else (np.zeros(len(t_stop)) if t_stop.shape else np.array(0.0))

		for m,(BR, chain) in enumerate(zip(*self._get_branches(isotope))):
			lm = self._r_lm(units)*self._chain[chain,0]
			L = len(chain)
			for i in range(L):
				if i==L-1 and m>0:
					continue

				ip = self.isotopes[chain[i]]
				A0 = self.A0[ip] if _A_dict is None else _A_dict[ip]
				A_i = lm[-1]*(A0/lm[i])
				B_i = np.prod(lm[i:-1]*BR[i:-1])

				for j in range(i, len(chain)):
					K = np.arange(i, len(chain))
					d_lm = lm[K[K!=j]]-lm[j]
					C_j = np.prod(np.where(np.abs(d_lm)>1E-12, d_lm, 1E-12*np.sign(d_lm)))
					if lm[j]>1E-12:
						D += A_i*B_i*(np.exp(-lm[j]*t_start)-np.exp(-lm[j]*t_stop))/(lm[j]*C_j)
					else:
						D += A_i*B_i*(t_stop-t_start)/C_j

		return D*self._r_lm((self.units if units is None else units), True)

	@property
	def counts(self):
		return self._counts

	@counts.setter
	def counts(self, N_c):
		if N_c is not None:
			if type(N_c)==pd.DataFrame:
				self._counts = N_c.copy(deep=True)
				self._counts['isotope'] = [self._filter_name(ip) for ip in self._counts['isotope']]

			elif type(N_c)!=dict:
				N_c = np.asarray(N_c)
				self._counts = pd.DataFrame({'isotope':self.isotopes[0],
											'start':N_c[:,0],
											'stop':N_c[:,1],
											'counts':N_c[:,2],
											'unc_counts':N_c[:,3]})

			else:
				self._counts = pd.DataFrame({'isotope':[],'start':[],'stop':[],'counts':[],'unc_counts':[]})
				for ip in N_c:
					ct = np.array(N_c[ip])
					if len(ct.shape)==1:
						ct = np.array([ct])
					ct = pd.DataFrame({'isotope':self._filter_name(ip),
										'start':ct[:,0],
										'stop':ct[:,1],
										'counts':ct[:,2],
										'unc_counts':ct[:,3]})
					self._counts = pd.concat([self._counts, ct], ignore_index=True).reset_index(drop=True)

			self._counts['activity'] = [p['counts']*self.activity(p['isotope'], p['start'])/self.decays(p['isotope'], p['start'], p['stop']) for n,p in self._counts.iterrows()]
			self._counts['unc_activity'] = self._counts['unc_counts']*self.counts['activity']/self._counts['counts']
			
		
	def get_counts(self, spectra, EoB, peak_data=None):
		"""Retrieves the number of measured decays

		Takes the number of measured decays from one of the following: a list of spectra,
		a file with peak data, or a pandas DataFrame with peak data.

		Parameters
		----------
		spectra : list or str
			List of ci.Spectrum objects, or str of spectra filenames.  If list of str, 
			peak_data **must** be specified.  In this case the filenames must be
			an exact match of the filenames in `peak_data`.  If spectra is a str,
			it is assumed to be a regex match for the filenames in `peak_data`.
		
		EoB : str or datetime.datetime
			Date/time of end-of-bombardment (t=0).  Must be a datetime object or
			a string in the format '%m/%d/%Y %H:%M:%S'.  This is used to calculate
			the decay time for the count.

		peak_data : str or pd.DataFrame, optional
			Either a file path to a file that was created using
			`ci.Spectrum.saveas()` or a DataFrame with the same 
			structure as `ci.Spectrum.peaks`.

		Examples
		--------
		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> sp.isotopes = ['152EU']
		>>> sp.saveas('test_spec.json')

		>>> dc = ci.DecayChain('152EU', A0=3.7E3, units='h')
		>>> dc.get_counts([sp], EoB='01/01/2016 08:39:08')
		>>> dc.get_counts(['eu_calib_7cm.Spe'], EoB='01/01/2016 08:39:08', peak_data='test_spec.json')
		>>> print(dc.counts)

		"""

		counts = []
		if type(EoB)==str:
			EoB = dtm.datetime.strptime(EoB, '%m/%d/%Y %H:%M:%S')

		if peak_data is not None:
			if type(peak_data)==str:
				if peak_data.endswith('.json'):
					peak_data = pd.read_json(peak_data, orient='records')
				elif peak_data.endswith('.csv'):
					peak_data = pd.read_csv(peak_data, header=0)
				elif peak_data.endswith('.db'):
					peak_data = pd.read_sql('SELECT * FROM peaks', _get_connection(peak_data))

			else:
				peak_data = pd.DataFrame(peak_data)


		if type(spectra)==str and peak_data is not None:
			df = peak_data['filename']
			spectra = list(set(map(str, df[df.str.contains(spectra)].to_list())))


		for sp in spectra:
			if type(sp)==str:
				if peak_data is not None:

					df = peak_data[peak_data['filename']==sp]
					df['isotope'] = [self._filter_name(i) for i in df['isotope']]
					df = df[df['isotope'].isin(self.isotopes)]

					if len(df):
						start_time = df.iloc[0]['start_time']
						if type(start_time)==str or type(start_time)==unicode:
							start_time = dtm.datetime.strptime(start_time, '%m/%d/%Y %H:%M:%S')
						start = (start_time-EoB).total_seconds()*self._r_lm('s')
						stop = start+(df.iloc[0]['real_time']*self._r_lm('s'))
				else:
					raise ValueError('peak_data must be specified if type(spectra)==str')
			else:
				if peak_data is not None:
					df = peak_data[peak_data['filename']==sp.filename]
				else:
					df = sp.peaks.copy()

				df['isotope'] = [self._filter_name(i) for i in df['isotope']]
				df = df[df['isotope'].isin(self.isotopes)]

				if len(df):
					start = (sp.start_time-EoB).total_seconds()*self._r_lm('s')
					stop = start+(sp.real_time*self._r_lm('s'))
			if len(df):
				counts.append(pd.DataFrame({'isotope':df['isotope'], 'start':start, 'stop':stop, 'counts':df['decays'], 'unc_counts':df['unc_decays']}))

		self.counts = pd.concat(counts, sort=True, ignore_index=True).sort_values(by=['start']).reset_index(drop=True)


	@property	
	def R_avg(self):
		df = []
		for ip in np.unique(self.R['isotope']):
			time = np.insert(np.unique(self.R['time']), 0, [0.0])
			df.append({'isotope':ip, 'R_avg':np.average(self.R[self.R['isotope']==ip]['R'], weights=time[1:]-time[:-1])})
		return pd.DataFrame(df)
		
	def fit_R(self, max_error=0.4, min_counts=1):
		"""Fit the production rate to count data

		Fits a scalar multiplier to the production rate (as a function of time) for
		each isotope specified in self.R.  The fit minimizes to the number of
		measured decays (self.counts) as a function of time, rather than the 
		activity, because the activity at each time point may be sensitive to
		the shape of the decay curve.

		Parameters
		----------
		max_error : float, optional
			The maximum relative error of a count datum to include in the fit. E.g. 0.25=25%
			(ony points with less than 25% error will be shown). Default, 0.4 (40%).

		min_counts : float or int, optional
			The minimum number of counts (decays) for a datum in self.counts to be included
			in the fit. Default, 1.

		Returns
		-------
		isotopes : list
			List of isotopes where R>0.  Same indices as fit. (i.e. isotope[0] corresponds
			to fit[0] and cov[0][0].)

		fit : np.ndarray
			The fitted time-averaged production rate for each isotope where R>0.

		cov : np.ndarray
			Covariance matrix on the fit.

		Examples
		--------
		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> sp.isotopes = ['152EU']

		>>> dc = ci.DecayChain('152EU', R=[[3E5, 36.0]], units='d')
		>>> dc.get_counts([sp], EoB='01/01/2016 08:39:08')
		>>> print(dc.fit_R())
		(array(['152EUg'], dtype=object), array([1291584.51735774]), array([[1.67412376e+09]]))

		"""

		if self.R is None:
			raise ValueError('Cannot fit R: R=0.')

		X = []
		R_isotopes = [i for i in self.isotopes if i in pd.unique(self.R['isotope'])]
		time = np.insert(np.unique(self.R['time']), 0, [0.0])
		filter_counts = self.counts[(self.counts['counts']>min_counts)&(self.counts['unc_counts']<max_error*self.counts['counts'])]

		for ip in R_isotopes:
			A0 = {p:0.0 for p in self.A0}
			for n,dt in enumerate(time[1:]-time[:-1]):
				_R_dict = {ip:self.R[self.R['isotope']==ip].iloc[n]['R']}
				A0 = {p:self.activity(p, dt, _R_dict=_R_dict, _A_dict=A0) for p in self.A0}

			X.append([self.decays(c['isotope'], c['start'], c['stop'], _A_dict=A0) for n,c in filter_counts.iterrows()])

		X = np.array(X)
		Y = filter_counts['counts'].to_numpy()
		dY = filter_counts['unc_counts'].to_numpy()

		
		wh = np.array([np.all(x>0) for x in X.T])
		p0 = np.average(Y[wh]/X[:,wh], axis=1)
		p0 = np.where((p0>0)&(np.isfinite(p0)), p0, 1.0)

		func = lambda X_f, *R_f: np.dot(np.asarray(R_f), X_f)
		fit, cov = curve_fit(func, X, Y, sigma=dY, p0=p0, bounds=(0.0*p0, np.inf*p0))


		for n,ip in enumerate(R_isotopes):
			df_sub = self.R[self.R['isotope']==ip]
			self.R.loc[df_sub.index, 'R'] = df_sub['R']*fit[n]

		self.A0 = {i:0.0 for i in self.A0}
		for n,dt in enumerate(time[1:]-time[:-1]):
			_R_dict = {p:self.R[self.R['isotope']==p].iloc[n]['R'] for p in pd.unique(self.R['isotope'])}
			self.A0 = {p:self.activity(p, dt, _R_dict=_R_dict) for p in self.A0}

		self.counts = self.counts

		R_avg = self.R_avg
		R_norm = np.array([R_avg[R_avg['isotope']==i]['R_avg'].to_numpy()[0] for i in R_isotopes])
		if not np.any(np.isfinite(np.diag(cov))):
			cov = np.ones(cov.shape)*((np.average(dY/Y))*fit)**2
		return R_isotopes, R_norm, cov*(R_norm/fit)**2
		
	def fit_A0(self, max_error=0.4, min_counts=1):
		"""Fit the initial activity to count data

		Fits a scalar multiplier to the initial activity for
		each isotope specified in self.A0.  The fit minimizes to the number of
		measured decays (self.counts) as a function of time, rather than the 
		activity, because the activity at each time point may be sensitive to
		the shape of the decay curve.

		Parameters
		----------
		max_error : float, optional
			The maximum relative error of a count datum to include in the fit. E.g. 0.25=25%
			(ony points with less than 25% error will be shown). Default, 0.4 (40%).

		min_counts : float or int, optional
			The minimum number of counts (decays) for a datum in self.counts to be included
			in the fit. Default, 1.

		Returns
		-------
		isotopes : list
			List of isotopes where A0>0.  Same indices as fit. (i.e. isotope[0] corresponds
			to fit[0] and cov[0][0].)

		fit : np.ndarray
			The initial activity for each isotope where A0>0.

		cov : np.ndarray
			Covariance matrix on the fit.

		Examples
		--------
		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> sp.isotopes = ['152EU']

		>>> dc = ci.DecayChain('152EU', A0=3.7E4, units='d')
		>>> dc.get_counts([sp], EoB='01/01/2016 08:39:08')
		>>> print(dc.fit_A0())
		(['152EUg'], array([6501.93665952]), array([[42425.53832341]]))

		"""

		if self.R is not None:
			raise ValueError('Cannot fit A0 when R!=0.')

		X = []
		A0_isotopes = [i for i in self.isotopes if i in self.A0]
		filter_counts = self.counts[(self.counts['counts']>min_counts)&(self.counts['unc_counts']<max_error*self.counts['counts'])]
		for ip in A0_isotopes:
			A0 = {p:(self.A0[p] if p==ip else 0.0) for p in self.A0}
			X.append([self.decays(c['isotope'], c['start'], c['stop'], _A_dict=A0) for n,c in filter_counts.iterrows()])

		X = np.array(X)
		Y = filter_counts['counts'].to_numpy()
		dY = filter_counts['unc_counts'].to_numpy()

		func = lambda X_f, *R_f: np.dot(np.asarray(R_f), X_f)
		p0 = np.ones(len(X))
		fit, cov = curve_fit(func, X, Y, sigma=dY, p0=p0, bounds=(0.0*p0, np.inf*p0))

		for n,ip in enumerate(A0_isotopes):
			self.A0[ip] *= fit[n]

		self.counts = self.counts

		A_norm = np.array([self.A0[i] for i in A0_isotopes])
		return A0_isotopes, A_norm, cov*(A_norm/fit)**2
		
	def plot(self, time=None, max_plot=10, max_label=10, max_plot_error=0.4, max_plot_chi2=10, **kwargs):
		"""Plot the activities in the decay chain

		Plots the activities as a function of time for all radioactive
		isotopes in the decay chain.  Can plot along a specified time
		grid, else the time will be inferred from the half-life of the
		parent isotope, or any count information given to self.counts.

		Parameters
		----------
		time : array_like, optional
			Time grid along which to plot.  Units must be the same as the decay chain.

		max_plot : int, optional
			Maximum number of isotope activities to plot in the decay chain. Default, 10.

		max_label : int, optional
			Maximum number of isotope activities to label in the legend. Default, 10.

		max_plot_error : float, optional
			The maximum relative error of a count point to include on the plot. E.g. 0.25=25%
			(ony points with less than 25% error will be shown). Default, 0.4.

		max_plot_chi2 : float or int, optional
			Maximum chi^2 of a count point to include on the plot. Only points with a chi^2
			less than this value will be shown. Default, 10.

		Other Parameters
		----------------
		**kwargs
			Optional keyword arguments for plotting.  See the 
			plotting section of the curie API for a complete
			list of kwargs.

		Examples
		--------
		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> sp.isotopes = ['152EU']

		>>> dc = ci.DecayChain('152EU', R=[[3E5, 36.0]], units='d')
		>>> dc.get_counts([sp], EoB='01/01/2016 08:39:08')
		>>> dc.fit_R()
		>>> dc.plot()

		>>> dc = ci.DecayChain('99MO', A0=350E6, units='d')
		>>> dc.plot()

		"""

		f, ax = _init_plot(**kwargs)

		if time is not None:
			time = np.asarray(time)
		elif self.counts is None:
			time = np.linspace(0, 5.0*np.log(2)/self._chain[0,0], 1000)
		else:
			time = np.linspace(0, 1.25*self.counts['stop'].max(), 1000)

		if max_plot is None:
			max_plot = len(self.isotopes)

		ordr = int(np.floor(np.log10(np.average(self.activity(self.isotopes[0], time)))/3.0))
		lb_or = {-5:'f',-4:'p',-3:'n',-2:r'$\mu$',-1:'m',0:'',1:'k',2:'M',3:'G',4:'T',5:'E'}[ordr]
		mult = 10**(-3*ordr)

		if self.R is not None:
			A0 = {p:0.0 for p in self.A0}
			T = np.insert(np.unique(self.R['time']), 0, [0.0])
			T_grid = np.array([])
			A_grid = {p:np.array([]) for p in A0}

			for n,dt in enumerate(T[1:]-T[:-1]):
				_R_dict = {p:self.R[self.R['isotope']==p].iloc[n]['R'] for p in pd.unique(self.R['isotope'])}
				dT = np.linspace(0, dt, 50)
				T_grid = dT if n==0 else np.append(T_grid, T_grid[-1]+dT)
				A_grid = {p:np.append(A_grid[p], self.activity(p, dT, _R_dict=_R_dict, _A_dict=A0)) for p in A0}
				A0 = {p:A_grid[p][-1] for p in A0}


		plot_time = time if self.R is None else np.append(T_grid-T_grid[-1], time.copy())
		for n,istp in enumerate(self.isotopes):
			if self._chain[n,0]>1E-12 and n<max_plot:
				A = self.activity(istp, time)
				if self.R is not None:
					A = np.append(A_grid[istp], A)

				label = Isotope(istp).TeX if n<max_label else None
				line, = ax.plot(plot_time, A*mult, label=label)

				if self.counts is not None:
					df = self.counts[self.counts['isotope']==istp]
					if len(df):
						x, y, yerr = df['start'].to_numpy(), df['activity'].to_numpy(), df['unc_activity'].to_numpy()
						idx = np.where((max_plot_error*y>yerr)&(yerr>0.0)&(np.isfinite(yerr)))
						if len(x[idx]>0):
							x, y, yerr = x[idx], y[idx], yerr[idx]
						idx = np.where((self.activity(istp, x)-y)**2/yerr**2<max_plot_chi2)
						if len(x[idx]>0):
							x, y, yerr = x[idx], y[idx], yerr[idx]
					
						ax.errorbar(x, y*mult, yerr=yerr*mult, ls='None', marker='o', color=line.get_color(), label=None)



		ax.set_xlabel('Time ({})'.format(self.units))
		ax.set_ylabel('Activity ({}Bq)'.format(lb_or))
		ax.legend(loc=0)
		return _draw_plot(f, ax, **kwargs)
		


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import pandas as pd
import json

from scipy.optimize import curve_fit
from scipy.interpolate import interp1d

from .isotope import Isotope
from .data import _get_connection
from .plotting import _init_plot, _draw_plot, colormap




_MU_Be = [[1.000000e+00, 1.117770e+03],
		 [1.500000e+00, 3.324450e+02],
		 [2.000000e+00, 1.381950e+02],
		 [3.000000e+00, 3.934950e+01],
		 [4.000000e+00, 1.606725e+01],
		 [5.000000e+00, 8.082650e+00],
		 [6.000000e+00, 4.676800e+00],
		 [8.000000e+00, 2.079400e+00],
		 [1.000000e+01, 1.196210e+00],
		 [1.500000e+01, 5.679500e-01],
		 [2.000000e+01, 4.164350e-01],
		 [3.000000e+01, 3.315200e-01],
		 [4.000000e+01, 3.034000e-01],
		 [5.000000e+01, 2.874900e-01],
		 [6.000000e+01, 2.762050e-01],
		 [8.000000e+01, 2.591850e-01],
		 [1.000000e+02, 2.456800e-01],
		 [1.500000e+02, 2.201500e-01],
		 [2.000000e+02, 2.014650e-01],
		 [3.000000e+02, 1.750655e-01],
		 [4.000000e+02, 1.567320e-01],
		 [5.000000e+02, 1.431715e-01],
		 [6.000000e+02, 1.323675e-01],
		 [8.000000e+02, 1.162910e-01],
		 [1.000000e+03, 1.045620e-01],
		 [1.022000e+03, 1.034335e-01],
		 [1.250000e+03, 9.349900e-02],
		 [1.500000e+03, 8.506300e-02],
		 [2.000000e+03, 7.285300e-02],
		 [2.044000e+03, 7.198350e-02],
		 [3.000000e+03, 5.805300e-02],
		 [4.000000e+03, 4.928400e-02],
		 [5.000000e+03, 4.341950e-02],
		 [6.000000e+03, 3.923850e-02],
		 [7.000000e+03, 3.609350e-02],
		 [8.000000e+03, 3.365150e-02],
		 [9.000000e+03, 3.170900e-02],
		 [1.000000e+04, 3.009950e-02]]


_MU_Ge = [[1.1100000e+01, 1.0465018e+03, 4.0529322e-01, 1.0550186e+03],
		 [1.5000000e+01, 4.8119920e+02, 4.8167827e-01, 4.8694804e+02],
		 [2.0000000e+01, 2.2058512e+02, 5.4667210e-01, 2.2473706e+02],
		 [3.0000000e+01, 7.1062050e+01, 6.1800030e-01, 7.3723550e+01],
		 [4.0000000e+01, 3.1102289e+01, 6.5100290e-01, 3.3034538e+01],
		 [5.0000000e+01, 1.6203212e+01, 6.6431040e-01, 1.7752205e+01],
		 [6.0000000e+01, 9.4483250e+00, 6.6750420e-01, 1.0768429e+01],
		 [8.0000000e+01, 4.0002345e+00, 6.5951970e-01, 5.0573823e+00],
		 [1.0000000e+02, 2.0413705e+00, 6.4408300e-01, 2.9542650e+00],
		 [1.5000000e+02, 5.9883750e-01, 5.9883750e-01, 1.3259593e+00],
		 [2.0000000e+02, 2.5188436e-01, 5.5785040e-01, 8.8415030e-01],
		 [3.0000000e+02, 7.6491510e-02, 4.9136613e-01, 6.0203130e-01],
		 [4.0000000e+02, 3.4131076e-02, 4.4282037e-01, 4.9647621e-01],
		 [5.0000000e+02, 1.8848743e-02, 4.0561260e-01, 4.3712476e-01],
		 [6.0000000e+02, 1.1912874e-02, 3.7596349e-01, 3.9666996e-01],
		 [8.0000000e+02, 6.0895120e-03, 3.3098414e-01, 3.4205598e-01],
		 [1.0000000e+03, 3.7899760e-03, 2.9787508e-01, 3.0484821e-01],
		 [1.0220000e+03, 3.6015418e-03, 2.9473451e-01, 3.0138826e-01],
		 [1.2500000e+03, 2.4336756e-03, 2.6657584e-01, 2.7152623e-01],
		 [1.5000000e+03, 1.7517993e-03, 2.4235619e-01, 2.4789211e-01],
		 [2.0000000e+03, 1.0757783e-03, 2.0690501e-01, 2.1749778e-01],
		 [2.0440000e+03, 1.0390496e-03, 2.0434997e-01, 2.1552827e-01],
		 [3.0000000e+03, 5.8020700e-04, 1.6288380e-01, 1.8758252e-01],
		 [4.0000000e+03, 3.8905807e-04, 1.3584296e-01, 1.7432825e-01],
		 [5.0000000e+03, 2.9031642e-04, 1.1731892e-01, 1.6810034e-01],
		 [6.0000000e+03, 2.3069882e-04, 1.0369204e-01, 1.6538561e-01],
		 [7.0000000e+03, 1.9093601e-04, 9.3258960e-02, 1.6474685e-01],
		 [8.0000000e+03, 1.6267088e-04, 8.4848620e-02, 1.6522592e-01],
		 [9.0000000e+03, 1.4164503e-04, 7.8035180e-02, 1.6639698e-01],
		 [1.0000000e+04, 1.2535665e-04, 7.2286340e-02, 1.6799388e-01]]

_MU_Be, _MU_Ge = np.array(_MU_Be), np.array(_MU_Ge)
_MU_W = lambda E_keV: np.exp(interp1d(np.log(_MU_Be[:,0]), np.log(_MU_Be[:,1]), bounds_error=False, fill_value='extrapolate', kind='quadratic')(np.log(E_keV)))
_TAU = lambda E_keV: np.exp(interp1d(np.log(_MU_Ge[:,0]), np.log(_MU_Ge[:,1]), bounds_error=False, fill_value='extrapolate', kind='quadratic')(np.log(E_keV)))
_SIGMA = lambda E_keV: np.exp(interp1d(np.log(_MU_Ge[:,0]), np.log(_MU_Ge[:,2]), bounds_error=False, fill_value='extrapolate', kind='quadratic')(np.log(E_keV)))
_MU = lambda E_keV: np.exp(interp1d(np.log(_MU_Ge[:,0]), np.log(_MU_Ge[:,3]), bounds_error=False, fill_value='extrapolate', kind='quadratic')(np.log(E_keV)))


class Calibration(object):
	"""Calibration for HPGe spectra

	Provides methods for calculating and fitting energy, efficiency and resolution
	calibrations for HPGe spectra.  Each Spectrum class contains a Calibration, 
	which can be loaded from a file, or fit to calibration spectra from known sources.
	
	Parameters
	----------
	filename : str, optional
		Path to a .json file storing calibration data.  The .json file can be produced by
		calling `cb.saveas('example_calib.json')` after performing a calibration fit with
		the `cb.calibrate()` function.  This allows past calibrations to be recalled without
		needing to re-perform the calibration fits.

	Attributes
	----------
	engcal : np.ndarray
		Energy calibration parameters. length 2 or 3 array, depending on whether the calibration
		is linear or quadratic.

	effcal : np.ndarray
		Efficiency calibration parameters. length 3 or 5 array, depending on whether the efficiency
		fit includes a "dead-layer term".

	unc_effcal : np.ndarray
		Efficiency calibration covariance matrix. shape 3x3 or 5x5, depending on the length of effcal.

	rescal : np.ndarray
		Resolution calibration parameters.  length 2 array if resolution calibration is of the form
		R = a + b*chan (default), or length 1 if R = a*sqrt(chan).


	Examples
	--------
	>>> cb = Calibration()
	>>> cb.engcal = [0.0, 0.25, 0.001]
	>>> print(cb.engcal)
	[0.0 0.3]
	>>> print(cb.effcal)
	[0.02 8.3 2.1 1.66 0.4]
	>>> cb.saveas('test_calib.json')
	>>> cb = Calibration('test_calib.json')
	>>> print(cb.engcal)
	[0.0 0.25 0.001]

	"""

	def __init__(self, filename=None):
		self._engcal = np.array([0.0, 0.3])
		self._effcal = np.array([1.87867638e-02, 8.82671055e+01, 2.09673703e+00, 1.66193806e+00,
								 3.90089610e-01, 3.79361265e+00, 1.24058591e-02])
		self._unc_effcal = np.array([[ 1.58657336e-05, -8.87692037e-17,  3.38230753e-04,  2.29975399e-04,
									  -2.52612161e-04,  3.15625462e-03, -3.28997933e-06],
									 [-8.87692037e-17,  4.96869464e-28, -1.88956043e-15, -1.27662324e-15,
									   1.40932100e-15, -1.76678605e-14,  1.84808443e-17],
									 [ 3.38230753e-04, -1.88956043e-15,  7.87574389e-03,  5.80822792e-03,
									  -5.65223905e-03,  6.70142358e-02, -6.81029068e-05],
									 [ 2.29975399e-04, -1.27662324e-15,  5.80822792e-03,  4.99639027e-03,
									  -4.20485186e-03,  4.51280328e-02, -4.29072666e-05],
									 [-2.52612161e-04,  1.40932100e-15, -5.65223905e-03, -4.20485186e-03,
									   4.20699447e-03, -5.00275244e-02,  5.06228293e-05],
									 [ 3.15625462e-03, -1.76678605e-14,  6.70142358e-02,  4.51280328e-02,
									  -5.00275244e-02,  6.28286136e-01, -6.57808573e-04],
									 [-3.28997933e-06,  1.84808443e-17, -6.81029068e-05, -4.29072666e-05,
									   5.06228293e-05, -6.57808573e-04,  7.23950095e-07]])
		self._rescal = np.array([2.0, 4E-4])
		self._calib_data = {}

		if filename is not None:
			if filename.endswith('.json'):
				with open(filename) as f:
					js = json.load(f)
					if 'engcal' in js:
						self.engcal = js['engcal']
					if 'rescal' in js:
						self.rescal = js['rescal']
					if 'effcal' in js:
						self.effcal = js['effcal']
					if 'unc_effcal' in js:
						self.unc_effcal = js['unc_effcal']
					if '_calib_data' in js:
						for c in ['engcal','rescal','effcal']:
							if c in js['_calib_data']:
								self._calib_data[c] = {str(i):np.array(js['_calib_data'][c][i]) for i in js['_calib_data'][c]}


	@property
	def engcal(self):
		return self._engcal

	@engcal.setter
	def engcal(self, cal):
		self._engcal = np.asarray(cal)

	@property
	def effcal(self):
		return self._effcal

	@effcal.setter
	def effcal(self, cal):
		self._effcal = np.asarray(cal)

	@property
	def unc_effcal(self):
		return self._unc_effcal

	@unc_effcal.setter
	def unc_effcal(self, cal):
		self._unc_effcal = np.asarray(cal)

	@property
	def rescal(self):
		return self._rescal

	@rescal.setter
	def rescal(self, cal):
		self._rescal = np.asarray(cal)


	def eng(self, channel, engcal=None):
		"""Energy calibration function

		Returns the calculated energy given an input array of
		channel numbers.  The `engcal` can be supplied, or
		if `engcal=None` the Calibration object's energy
		calibration is used (cb.engcal).

		Parameters
		----------
		channel : array_like
			Spectrum channel number.  The maxi_MUm channel number should
			be the length of the spectrum.

		engcal : array_like, optional
			Optional energy calibration. If a length 2 array, calibration
			will be engcal[0] + engcal[1]*channel.  If length 3, then
			engcal[0] + engcal[1]*channel + engcal[2]*channel**2

		Returns
		-------
		energy: np.ndarray
			Calibrated energy corresponding to the given channels.

		Examples
		--------
		>>> cb = ci.Calibration()
		>>> print(cb.engcal)
		[0.  0.3]
		>>> print(cb.eng(np.arange(10)))
		[0.  0.3 0.6 0.9 1.2 1.5 1.8 2.1 2.4 2.7]
		>>> cb.engcal = [0.1, 0.2, 0.003]
		>>> print(cb.eng(np.arange(10)))
		[0.1   0.303 0.512 0.727 0.948 1.175 1.408 1.647 1.892 2.143]

		"""

		channel = np.asarray(channel)

		if engcal is None:
			engcal = self.engcal

		if len(engcal)==2:
			return engcal[0] + engcal[1]*channel

		elif len(engcal)==3:
			return engcal[0] + engcal[1]*channel + engcal[2]*channel**2
		
	def eff(self, energy, effcal=None):
		"""Efficiency calibration function

		Returns the calculated (absolute) efficiency given an input array of 
		energies. The `effcal` can be supplied, or if `effcal=None`, the
		calibration object's internal efficiency calibration (cb.effcal)
		is used.

		The functional form of the efficiency used is a modified version of the
		semi-empirical formula proposed by Vidmar (2001):
		eff(E) = c[0]*(1.0-exp(-mu(eng)*c[1]))*(tau(eng)+sigma(eng)*(1.0-exp(-(mu(eng)*c[3])**c[2]))*c[4])/mu(eng) if the effcal is length 5 or
		eff(E) = c[0]*exp(-mu_w(eng)*c[5])*exp(-mu(eng)*c[6])*(1.0-exp(-mu(eng)*c[1]))*(tau(eng)+sigma(eng)*(1.0-exp(-(mu(eng)*c[3])**c[2]))*c[4])/mu(eng)
		if the effcal is length 7.

		Parameters
		----------
		energy : array_like
			Peak energy in keV.

		effcal : array_like, optional
			Efficiency calibration parameters. length 5 or 7 array, depending on whether the efficiency
			fit includes the low-energy components.

		Returns
		-------
		efficiency : np.ndarray
			Absolute efficiency at the given energies.

		Examples
		--------
		>>> cb = ci.Calibration()
		>>> print(cb.effcal)
		[0.02 8.3 2.1 1.66 0.4]
		>>> print(cb.eff(50*np.arange(1,10)))
		[0.04152215 0.06893742 0.07756411 0.07583971 0.07013108 0.06365222
		 0.05762826 0.05236502 0.0478359 ]

		"""

		energy = np.asarray(energy)

		if effcal is None:
			effcal = self.effcal


		if len(effcal)==5:
			sa, l, alpha, l0, kappa = tuple(effcal)
			return sa*(1.0-np.exp(-_MU(energy)*l))*(_TAU(energy)+_SIGMA(energy)*(1.0-np.exp(-(_MU(energy)*l0)**alpha))*kappa)/_MU(energy)

		else:
			sa, l, alpha, l0, kappa, w, d = tuple(effcal)
			return sa*np.exp(-_MU_W(energy)*w)*np.exp(-_MU(energy)*d)*(1.0-np.exp(-_MU(energy)*l))*(_TAU(energy)+_SIGMA(energy)*(1.0-np.exp(-(_MU(energy)*l0)**alpha))*kappa)/_MU(energy)

		
	def unc_eff(self, energy, effcal=None, unc_effcal=None):
		"""Uncertainty in the efficiency

		Returns the calculated uncertainty in efficiency for an input
		array of energies.  If `effcal` or `unc_effcal` are not none,
		they are used instead of the calibration object's internal
		values. (cb.unc_effcal)  unc_effcal _MUst be a covariance matrix of the
		same dimension as effcal.

		Parameters
		----------
		energy : array_like
			Peak energy in keV.

		effcal : array_like, optional
			Efficiency calibration parameters. length 5 or 7 array, depending on whether the efficiency
			fit includes the low-energy components.

		unc_effcal : array_like, optional
			Efficiency calibration covariance matrix. shape 5x5 or 7x7, depending on the length of effcal.

		Returns
		-------
		unc_efficiency : np.ndarray
			Absolute uncertainty in efficiency for the given energies.

		Examples
		--------
		>>> cb = ci.Calibration()
		>>> print(cb.effcal)
		[0.02 8.3 2.1 1.66 0.4]
		>>> print(cb.unc_effcal)
		[[ 5.038e-02  3.266e-02 -2.151e-02 -4.869e-05 -7.748e-03]
		 [ 3.266e-02  2.144e-02 -1.416e-02 -3.416e-05 -4.137e-03]
		 [-2.151e-02 -1.416e-02  9.367e-03  2.294e-05  2.569e-03]
		 [-4.869e-05 -3.416e-05  2.294e-05  5.411e-07 -1.165e-04]
		 [-7.748e-03 -4.137e-03  2.569e-03 -1.165e-04  3.332e-02]]
		>>> print(cb.eff(50*np.arange(1,10)))
		[0.04152215 0.06893742 0.07756411 0.07583971 0.07013108 0.06365222
		 0.05762826 0.05236502 0.0478359 ]
		>>> print(cb.unc_eff(50*np.arange(1,10)))
		[0.00453714 0.00795861 0.00705099 0.00514765 0.00453579 0.00433504
		 0.00394935 0.00347228 0.00304041]

		"""

		energy = np.asarray(energy)

		if effcal is None or unc_effcal is None:
			effcal, unc_effcal = self.effcal, self.unc_effcal

		eps = np.abs(1E-2*effcal)
		var = np.zeros(len(energy)) if energy.shape else 0.0

		for n in range(len(effcal)):
			for m in range(n, len(effcal)):

				if not np.isfinite(unc_effcal[n][m]):
					return np.inf*(var+1.0)

				c_n, c_m = np.copy(effcal), np.copy(effcal)
				c_n[n], c_m[m] = c_n[n]+eps[n], c_m[m]+eps[m]

				par_n = (self.eff(energy, c_n)-self.eff(energy, effcal))/eps[n]
				par_m = (self.eff(energy, c_m)-self.eff(energy, effcal))/eps[m]

				var += unc_effcal[n][m]*par_n*par_m*(2.0 if n!=m else 1.0)

		return np.sqrt(var)
		
	def res(self, channel, rescal=None):
		"""Resolution calibration

		Calculates the expected 1-sigma peak widths for a given input array
		of channel numbers.  If `rescal` is given, it is used instead of
		the calibration object's internal value (cb.rescal).

		Parameters
		----------
		channel : array_like
			Spectrum channel number.  The maxi_MUm channel number should
			be the length of the spectrum.

		rescal : array_like, optional
			Resolution calibration parameters.  length 2 array if resolution calibration is of the form
			R = a + b*chan (default), or length 1 if R = a*sqrt(chan).

		Returns
		-------
		resolution : np.ndarray
			Calculated 1-sigma width of the peaks given the input channel numbers.

		Examples
		--------
		>>> cb = ci.Calibration()
		>>> print(cb.rescal)
		[2.e+00 4.e-04]
		>>> print(cb.res(100*np.arange(1,10)))
		[2.04 2.08 2.12 2.16 2.2  2.24 2.28 2.32 2.36]

		"""

		channel = np.asarray(channel)

		if rescal is None:
			rescal = self.rescal

		if len(rescal)==1:
			return rescal[0]*np.sqrt(channel)

		elif len(rescal)==2:
			return rescal[0] + rescal[1]*channel
		
	def map_channel(self, energy, engcal=None):
		"""Energy to channel calibration

		Calculates the spectrum channel number corresponding to a given
		energy array.  This should return the inverse of the energy calibration, but
		as an integer-type channel number.

		Parameters
		----------
		energy : array_like
			Peak energy in keV

		engcal : array_like, optional
			Energy calibration parameters. length 2 or 3 array, depending on whether the calibration
			is linear or quadratic.

		Returns
		-------
		channel : np.ndarray
			Calculated channel number given the input energy array

		Examples
		--------
		>>> cb = ci.Calibration()
		>>> print(cb.engcal)
		[0.  0.3]
		>>> print(cb.map_channel(300))
		1000
		>>> print(cb.eng(cb.map_channel(300)))
		300.0

		"""

		energy = np.asarray(energy)

		if engcal is None:
			engcal = self.engcal

		if len(engcal)==3:
			if engcal[2]!=0.0:
				return np.array(np.rint(0.5*(np.sqrt(engcal[1]**2-4.0*engcal[2]*(engcal[0]-energy))-engcal[1])/engcal[2]), dtype=np.int32)

		return np.array(np.rint((energy-engcal[0])/engcal[1]), dtype=np.int32)
		
	def calibrate(self, spectra, sources):
		"""Generate calibration parameters from spectra

		Performs an energy, resolution and efficiency calibration on peak fits
		to a given list of spectra.  Reference activities _MUst be given for the
		efficiency calibration.  Spectra are allowed to have isotopes that are
		not in `sources`, but these will not be included in the efficiency calibration.

		Parameters
		----------
		spectra : list of sp.Spectrum
			List of calibration spectra.  Must have sp.isotopes defined, and matching
			the isotopes given in `sources`.

		sources : str, list, dict or pd.DataFrame
			Datatype or (if str) file that can be converted into a
			pandas DataFrame.  Required keys are 'isotope', 'A0' (reference activity),
			and 'ref_date' (reference date).

		Examples
		--------
		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> sp.isotopes = ['152EU']

		>>> cb = ci.Calibration()
		>>> cb.calibrate([sp], sources=[{'isotope':'152EU', 'A0':3.5E4, 'ref_date':'01/01/2009 12:00:00'}])
		>>> print(cb.effcal)
		[4.33742771 2.28579733 0.15337749]
		>>> cb.plot()

		"""

		if type(sources)==str:
			if sources.endswith('.json'):
				sources = pd.DataFrame(json.loads(open(sources).read()))
			elif sources.endswith('.csv'):
				sources = pd.read_csv(sources, header=0).fillna(method='ffill')
			elif sources.endswith('.db'):
				sources = pd.read_sql('SELECT * FROM sources', _get_connection(sources))
		else:
			if type(sources)==dict:
				if type(sources['A0'])==float:
					sources['A0'] = [sources['A0']]
			sources = pd.DataFrame(sources)
		sources['ref_date'] = pd.to_datetime(sources['ref_date'], format='%m/%d/%Y %H:%M:%S')

		src_itps = sources['isotope'].to_list()
		lm = {ip:Isotope(ip).dc for ip in src_itps}
		unc_lm = {ip:Isotope(ip).decay_const(unc=True)[1] for ip in src_itps}

		specs = []
		for n,sp in enumerate(spectra):
			if type(sp)==str:
				from .spectrum import Spectrum
				spec = Spectrum(sp)
				spec.isotopes = sources['isotope'].to_list()
				specs.append(spec)
			else:
				specs.append(sp)
		spectra = specs

		cb_dat = []
		for sp in spectra:
			if sp._fits is None:
				sp.fit_peaks()

			cfg, ix = sp.fit_config, -1

			for ft in sp._fits:
				f, u = ft['fit'], ft['unc']
				B = {'snip':0, 'constant':1, 'linear':2, 'quadratic':3}[cfg['bg'].lower()]
				L = 3+2*int(cfg['skew_fit'])+int(cfg['step_fit'])

				for n in range(int((len(f)-B)/L)):
					ix += 1
					pk = sp.peaks.loc[ix, :]
					if pk['isotope'] not in src_itps:
						continue

					j = B+n*L+1
					mu, sig = f[j], f[j+1]
					unc_mu, unc_sig = np.sqrt(u[j][j]), np.sqrt(u[j+1][j+1])
					eng = pk['energy']

					src = sources[sources['isotope']==pk['isotope']].reset_index(drop=True)
					rd, A0 = src.loc[0, 'ref_date'], src.loc[0, 'A0']
					unc_A0 = src.loc[0, 'unc_A0'] if 'unc_A0' in src.columns else 0.0
					td = (sp.start_time-rd).total_seconds()
					dc = lm[pk['isotope']]

					if sp._atten_corr is None:
						corr = sp._geom_corr
					else:
						corr = sp._geom_corr*sp._atten_corr(pk['energy'])

					eff = pk['counts']*dc/(corr*(1.0-np.exp(-dc*sp.real_time))*np.exp(-dc*td)*pk['intensity']*A0*(sp.live_time/sp.real_time))
					unc_eff = eff*np.sqrt((pk['unc_counts']/pk['counts'])**2+(pk['unc_intensity']/pk['intensity'])**2+(unc_lm[pk['isotope']]/dc)**2+(unc_A0/A0)**2)

					cb_dat.append([mu, eng, unc_mu, sig, unc_sig, eff, unc_eff])

		cb_dat = np.array(cb_dat)
		self._calib_data = {'engcal':{'channel':cb_dat[:,0], 'energy':cb_dat[:,1], 'unc_channel':cb_dat[:,2]},
							'rescal':{'channel':cb_dat[:,0], 'width':cb_dat[:,3], 'unc_width':cb_dat[:,4]},
							'effcal':{'energy':cb_dat[:,1], 'efficiency':cb_dat[:,5], 'unc_efficiency':cb_dat[:,6]}}

		x, y, yerr = self._calib_data['engcal']['channel'], self._calib_data['engcal']['energy'], self.eng(self._calib_data['engcal']['unc_channel'])
		idx = np.where((0.25*y>yerr)&(yerr>0.0)&(np.isfinite(yerr)))
		x, y, yerr = x[idx], y[idx], yerr[idx]
		fn = lambda x, *A: self.eng(x, A)
		fit, unc = curve_fit(fn, x, y, sigma=yerr, p0=spectra[0].cb.engcal)
		self._calib_data['engcal'] = {'channel':x, 'energy':y, 'unc_channel':yerr, 'fit':fit, 'unc':unc}
		self.engcal = fit

		x, y, yerr = self._calib_data['rescal']['channel'], self._calib_data['rescal']['width'], self._calib_data['rescal']['unc_width']
		idx = np.where((0.33*y>yerr)&(yerr>0.0)&(np.isfinite(yerr)))
		x, y, yerr = x[idx], y[idx], yerr[idx]
		fn = lambda x, *A: self.res(x, A)
		fit, unc = curve_fit(fn, x, y, sigma=yerr, p0=spectra[0].cb.rescal)
		idx = np.where((self.res(x, fit)-y)**2/yerr**2 < 10.0)
		self._calib_data['rescal'] = {'channel':x[idx], 'width':y[idx], 'unc_width':yerr[idx], 'fit':fit, 'unc':unc}
		self.rescal = fit

		x, y, yerr = self._calib_data['effcal']['energy'], self._calib_data['effcal']['efficiency'], self._calib_data['effcal']['unc_efficiency']
		idx = np.where((0.33*y>yerr)&(yerr>0.0)&(np.isfinite(yerr)))
		x, y, yerr = x[idx], y[idx], yerr[idx]
		fn = lambda x, *A: self.eff(x, A)

		p0 = spectra[0].cb.effcal
		p0 = p0.tolist() if len(p0)==7 else p0.tolist()+[0.5, 0.001]
		p0[0] = max([min([p0[0]*np.average(y/self.eff(x, p0), weights=(self.eff(x, p0)/yerr)**2),4.99]),0.0001])
		bounds = ([0.0, 0.0, 0.1, 0.0, 0.0, 0.001, 1E-12], [12, 100, 6, 50, 6, 6, 0.2])

		if any([sp.fit_config['xrays'] for sp in spectra]):
			try:
				fit5, unc5 = curve_fit(fn, x, y, sigma=yerr, p0=p0[:5], bounds=(bounds[0][:5], bounds[1][:5]))
				fit7, unc7 = curve_fit(fn, x, y, sigma=yerr, p0=fit5.tolist()+p0[5:], bounds=bounds)
				
				chi7 = np.sum((y-self.eff(x, fit7))**2/yerr**2)
				chi5 = np.sum((y-self.eff(x, fit5))**2/yerr**2)
				## Invert to find which is closer to one
				chi7 = chi7 if chi7>1.0 else 1.0/chi7
				chi5 = chi5 if chi5>1.0 else 1.0/chi5
				fit, unc = (fit5, unc5) if chi5<=chi7 else (fit7, unc7)
			except:
				fit, unc = curve_fit(fn, x, y, sigma=yerr, p0=p0[:5], bounds=(bounds[0][:5], bounds[1][:5]))

		else:
			fit, unc = curve_fit(fn, x, y, sigma=yerr, p0=p0[:5], bounds=(bounds[0][:5], bounds[1][:5]))

		idx = np.where((self.eff(x, fit)-y)**2/yerr**2 < 10.0)
		self._calib_data['effcal'] = {'energy':x[idx], 'efficiency':y[idx], 'unc_efficiency':yerr[idx], 'fit':fit, 'unc':unc}
		self.effcal = fit
		self.unc_effcal = unc

		for sp in spectra:
			sp.cb.engcal = self._calib_data['engcal']['fit']
			sp.cb.rescal = self._calib_data['rescal']['fit']
			sp.cb.effcal = self._calib_data['effcal']['fit']
			sp.cb.unc_effcal = self._calib_data['effcal']['unc']
			sp._peaks, sp._fits = None, None



		
	def saveas(self, filename):
		"""Save the calibration as a .json file

		Saves the energy, resolution and efficiency calibration to a .json
		file, which can be recalled by a new calibration object by passing
		the 'filename' keyword to Calibration() upon construction.

		Parameters
		----------
		filename : str
			Complete filename to save the calibration. Must end in '.json'.

		Examples
		--------
		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> sp.isotopes = ['152EU']

		>>> cb = ci.Calibration()
		>>> cb.calibrate([sp], sources=[{'isotope':'152EU', 'A0':3.5E4, 'ref_date':'01/01/2009 12:00:00'}])
		>>> print(cb.effcal)
		[4.33742771 2.28579733 0.15337749]
		>>> cb.saveas('example_calib.json')

		>>> cb = ci.Calibration('example_calib.json')
		>>> print(cb.effcal)
		[4.33742771 2.28579733 0.15337749]

		"""

		if any([filename.endswith(e) for e in ['.png','.pdf','.eps','.pgf','.ps','.raw','.rgba','.svg','.svgz']]):
				self.plot(saveas=filename, show=False)

		if filename.endswith('.json'):
			js = {'engcal':self.engcal.tolist(),
				  'effcal':self.effcal.tolist(),
				  'unc_effcal':self.unc_effcal.tolist(),
				  'rescal':self.rescal.tolist()}
			if self._calib_data:
				js['_calib_data'] = {}
				for cl in ['engcal','rescal','effcal']:
					if cl in self._calib_data:
						js['_calib_data'][cl] = {i:self._calib_data[cl][i].tolist() for i in self._calib_data[cl]}

			with open(filename, 'w') as f:
				json.dump(js, f, indent=4)

		
	def plot_engcal(self, **kwargs):
		"""Plot the energy calibration

		Draws the energy calibration, with measurements from peak fit data if available.

		Other Parameters
		----------------
		**kwargs
			Optional keyword arguments for plotting.  See the 
			plotting section of the curie API for a complete
			list of kwargs.

		Examples
		--------
		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> sp.isotopes = ['152EU']

		>>> cb = ci.Calibration()
		>>> cb.calibrate([sp], sources=[{'isotope':'152EU', 'A0':3.5E4, 'ref_date':'01/01/2009 12:00:00'}])
		>>> cb.plot_engcal()

		"""

		f, ax = _init_plot(**kwargs)

		if 'engcal' in self._calib_data:
			d = self._calib_data['engcal']
			ax.errorbar(d['energy'], d['channel'], yerr=d['unc_channel'], ls='None', marker='o')
			x = np.arange(min(d['energy']), max(d['energy']), 0.1)
			ax.plot(x, self.map_channel(x, d['fit']))

		else:
			x = np.arange(20, 3000, 0.1)
			ax.plot(x, self.map_channel(x))

		ax.set_xlabel('Energy (keV)')
		ax.set_ylabel('ADC Channel')

		return _draw_plot(f, ax, **kwargs)
		
	def plot_rescal(self, **kwargs):
		"""Plot the resolution calibration

		Draws the resolution calibration, with measurements from peak fit data if available.

		Other Parameters
		----------------
		**kwargs
			Optional keyword arguments for plotting.  See the 
			plotting section of the curie API for a complete
			list of kwargs.

		Examples
		--------
		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> sp.isotopes = ['152EU']

		>>> cb = ci.Calibration()
		>>> cb.calibrate([sp], sources=[{'isotope':'152EU', 'A0':3.5E4, 'ref_date':'01/01/2009 12:00:00'}])
		>>> cb.plot_rescal()

		"""

		f, ax = _init_plot(**kwargs)

		if 'rescal' in self._calib_data:
			d = self._calib_data['rescal']
			ax.errorbar(d['channel'], d['width'], yerr=d['unc_width'], ls='None', marker='o')
			x = np.arange(min(d['channel']), max(d['channel']), 0.1)
			ax.plot(x, self.res(x, d['fit']))

		else:
			x = np.arange(0, 2**14, 0.1)
			ax.plot(x, self.res(x))

		ax.set_xlabel('ADC Channel')
		ax.set_ylabel('Peak Width')

		return _draw_plot(f, ax, **kwargs)
		
	def plot_effcal(self, **kwargs):
		"""Plot the efficiency calibration

		Draws the efficiency calibration, with measurements from peak fit data if available.

		Other Parameters
		----------------
		**kwargs
			Optional keyword arguments for plotting.  See the 
			plotting section of the curie API for a complete
			list of kwargs.

		Examples
		--------
		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> sp.isotopes = ['152EU']

		>>> cb = ci.Calibration()
		>>> cb.calibrate([sp], sources=[{'isotope':'152EU', 'A0':3.5E4, 'ref_date':'01/01/2009 12:00:00'}])
		>>> cb.plot_effcal()

		"""
		cm = colormap()
		cm_light = colormap(shade='light')

		f, ax = _init_plot(**kwargs)

		if 'effcal' in self._calib_data:
			d = self._calib_data['effcal']
			ax.errorbar(d['energy'], d['efficiency'], yerr=d['unc_efficiency'], ls='None', marker='o', color=cm['k'])
			x = np.arange(min(d['energy']), max(d['energy']), 0.1)
			ax.plot(x, self.eff(x, d['fit']), color=cm['k'])
			low = self.eff(x, d['fit'])-self.unc_eff(x, d['fit'], d['unc'])
			high = self.eff(x, d['fit'])+self.unc_eff(x, d['fit'], d['unc'])
			ax.fill_between(x, low, high, facecolor=cm_light['k'], alpha=0.5)

		else:
			x = np.arange(20, 3000, 0.1)
			ax.plot(x, self.eff(x), color=cm['k'])
			low = self.eff(x)-self.unc_eff(x)
			high = self.eff(x)+self.unc_eff(x)
			ax.fill_between(x, low, high, facecolor=cm_light['k'], alpha=0.5)

		ax.set_xlabel('Energy (keV)')
		ax.set_ylabel('Efficiency (abs.)')

		return _draw_plot(f, ax, **kwargs)
		
	def plot(self, **kwargs):
		"""Plots energy, resolution and efficiency calibrations

		Draws all three of energy, resolution and efficiency calibrations on a 
		single figure, and shows measured values from peak data, if available.

		Other Parameters
		----------------
		**kwargs
			Optional keyword arguments for plotting.  See the 
			plotting section of the curie API for a complete
			list of kwargs.

		Examples
		--------
		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> sp.isotopes = ['152EU']

		>>> cb = ci.Calibration()
		>>> cb.calibrate([sp], sources=[{'isotope':'152EU', 'A0':3.5E4, 'ref_date':'01/01/2009 12:00:00'}])
		>>> cb.plot()

		"""

		
		f, ax = _init_plot(_N_plots=3, figsize=(12.8, 4.8), **kwargs)

		f, ax[0] = self.plot_engcal(f=f, ax=ax[0], show=False, return_plot=True)
		ax[0].set_title('Energy Calibration')
		f, ax[1] = self.plot_rescal(f=f, ax=ax[1], show=False, return_plot=True)
		ax[1].set_title('Resolution Calibration')
		f, ax[2] = self.plot_effcal(f=f, ax=ax[2], show=False, return_plot=True)
		ax[2].set_title('Efficiency Calibration')

		return _draw_plot(f, ax, **kwargs)
		




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re
import os
import json
import numpy as np
import pandas as pd

from scipy.interpolate import interp1d

from .data import _get_connection
from .plotting import _init_plot, _draw_plot
from .element import ELEMENTS, Element

COMPOUND_LIST = list(map(str, pd.read_sql('SELECT compound FROM compounds', _get_connection('ziegler'))['compound']))



class Compound(object):
	"""Data and properties of atomic compounds

	The compound class provides the same functionality as the
	element class, but for compounds of atomic elements rather
	than the individual atomic elements.  The compound is described
	by a set of elements, a set of weights for each element, and a 
	density.  

	The weights can be either given as atom-weights, e.g.
	in H2O the atom weights are 0.67 for H and 0.33 for O, or as 
	mass-weights, e.g. brass is 0.33 Zn by weight and 0.67 Cu by
	weight.  Some preset compounds are available; their names can be
	found by printing `ci.COMPOUND_LIST`.
	
	Parameters
	----------
	compound : str
		The name of the compound.  If the compound is in `ci.COMPOUND_LIST`,
		the weights and density will take preset values, if not given.
		If the name of the compound is given in chemical notation, e.g. NaCl
		or H2O2, the atom-weights can be inferred if not given explicitly.  Note
		that full chemical notation is not supported, so Ca3(PO4)2 must be written
		as Ca3P2O8.  Decimal weights are supported, e.g. C0.5O is equivalent to CO2.

	weights : dict, str or pd.DataFrame, optional
		The weights of each element in the compound.  Multiple formats are supported.
		If weights is a dict, it must be formatted as {'el1':wt1, 'el2':wt2}, where
		atom-weights are positive, and mass-weights are negative. 

		If weights is a pandas DataFrame, it must contain an 'element' column, 
		and one of 'weight', 'atom_weight', or 'mass_weight'.  If 'weight' is the column 
		given, the convention of positive atom-weights and negative mass-weights is followed.  

		If weights is a str, it can either be formatted as 'el1:wt1, el2:wt2', or it can be 
		a path to a .csv, .json or .db file.  These files must contain the same information 
		as the DataFrame option, and can contain weights for multiple compounds, if 'compound' 
		is one of the columns/keys.
		
		If a .json file, it must follow the 'records' formatting convention (see pandas docs).

	density : float, optional
		Density of the compound in g/cm^3.  The density is required for the cm.attenuation(),
		cm.S(), cm.range() and cm.plot_range() functions, but is an optional argument in each
		of those functions if not provided at construction.  Can also be specified by using
		a 'density' column/key in the file/DataFrame for weights.

	Attributes
	----------
	name : str
		The name of the compound.

	weights : pd.DataFrame
		The weights for each element in the compound.  DataFrame columns are
		'element', 'Z', 'mass_weight', 'atom_weight'.

	density : float
		Density of the compound in g/cm^3. The density is used in
		calculations of charged particle dEdx and photon attenuation, so
		if the density was not explicitly given at construction,
		you can assign a new density using `cm.density = new_density` if
		needed, or using the `density` keyword in either of those functions.

	elements : list of ci.Element
		Elements in the compound.

	mass_coeff : pd.DataFrame
		Table of mass-attenuation coefficients as a function of photon
		energy, from the NIST XCOM database.  Energies are in keV, and
		mass-attenuation coefficients, or mu/rho, are given in cm^2/g.
		DataFrame columns are 'energy', 'mu' and 'mu_en' for the 
		mass-energy absorption coefficient.

	Examples
	--------
	>>> print('Silicone' in ci.COMPOUND_LIST)
	True
	>>> cm = ci.Compound('Silicone') # preset compound
	>>> print(list(map(str, cm.elements)))
	['H', 'C', 'O', 'Si']
	>>> cm = ci.Compound('H2O', density=1.0)
	print(cm.weights)
	  element  Z  atom_weight  mass_weight
	0       H  1     0.666667     0.111907
	1       O  8     0.333333     0.888093
	>>> cm = ci.Compound('Brass', weights={'Zn':-33,'Cu':-66})
	>>> print(cm.weights)
	  element   Z  atom_weight  mass_weight
	0      Zn  30     0.327041     0.333333
	1      Cu  29     0.672959     0.666667
	>>> cm.saveas('brass.csv')

	"""

	def __init__(self, compound, weights=None, density=None):
		self.name = compound
		self.density = None

		if compound in COMPOUND_LIST:
			df = pd.read_sql('SELECT * FROM compounds WHERE compound="{}"'.format(compound), _get_connection('ziegler'))
			self.density = df['density'][0]

			if weights is None:
				wts = df['weights'][0].split(',')
				elements = [str(i.split(':')[0]) for i in wts]
				atom_weights = np.array([float(i.split(':')[1]) for i in wts])
				self._set_weights(elements, atom_weights=atom_weights)
			
		elif weights is None:
			elements = []
			for el_gp in [i for i in re.split('[0-9]+|\\.', compound) if i]:
				for s in [i for i in re.split('([A-Z])', el_gp) if i]:
					if s.upper()==s:
						elements.append(s)
					else:
						elements[-1] += s

			if all([e in ELEMENTS for e in elements]):
				wts = re.split('|'.join(sorted(elements, key=lambda i:-len(i))), compound)
				atom_weights = np.array([float(wts[n+1]) if wts[n+1] else 1.0 for n,e in enumerate(elements)])
				self._set_weights(elements, atom_weights=atom_weights)

		if weights is not None:

			if type(weights)==dict:
				elements = [e for e in weights]
				wts = np.array([weights[e] for e in elements], dtype=np.float64)
				weights = pd.DataFrame({'element':elements, 'weight':wts})

			elif type(weights)==str:
				if weights.endswith('.json'):
					weights = pd.read_json(weights, orient='records').fillna(method='ffill')
					weights.columns = map(str.lower, map(str, weights.columns))
					if 'compound' in weights.columns:
						weights = weights[weights['compound']==self.name]

				elif weights.endswith('.csv'):
					weights = pd.read_csv(weights, header=0).fillna(method='ffill')
					weights.columns = map(str.lower, map(str, weights.columns))
					if 'compound' in weights.columns:
						weights = weights[weights['compound']==self.name]

				elif weights.endswith('.db'):
					weights = pd.read_sql('SELECT * FROM compounds WHERE compound={}'.format(self.name), _get_connection(weights))
					weights.columns = map(str.lower, map(str, weights.columns))
					if 'compound' in weights.columns:
						weights = weights[weights['compound']==self.name]

				else:
					elements = [str(i.split(':')[0]).strip() for i in weights.split(',')]
					wts = np.array([float(i.split(':')[1].strip()) for i in weights.split(',')])

					if wts[0]>0:
						self._set_weights(elements, atom_weights=wts)
					else:
						self._set_weights(elements, mass_weights=np.abs(wts))

			if type(weights)==pd.DataFrame:
				weights.columns = map(str.lower, map(str, weights.columns))
				if 'density' in weights.columns:
					self.density = weights['density'].iloc[0]

				cols = ['element', 'Z', 'atom_weight', 'mass_weight']
				if all([i in weights.columns for i in cols]):
					self.weights = weights[cols]

				elif 'atom_weight' in weights.columns:
					self._set_weights(list(weights['element']), atom_weights=weights['atom_weight'].to_numpy())

				elif 'mass_weight' in weights.columns:
					self._set_weights(list(weights['element']), mass_weights=weights['mass_weight'].to_numpy())

				else:
					elements, wts = list(weights['element']), weights['weight'].to_numpy()

					if wts[0]>0:
						self._set_weights(elements, atom_weights=wts)
					else:
						self._set_weights(elements, mass_weights=np.abs(wts))


		if density is not None:
			self.density = density

		self.elements = [Element(el) for el in self.weights['element']]

		if self.density is None and len(self.weights)==1:
			self.density = self.elements[0].density

		E = np.unique(np.concatenate([el.mass_coeff['energy'].to_numpy() for el in self.elements]))
		mu = np.average([el.mu(E) for el in self.elements], weights=self.weights['mass_weight'], axis=0)
		mu_en = np.average([el.mu_en(E) for el in self.elements], weights=self.weights['mass_weight'], axis=0)
		self.mass_coeff = pd.DataFrame({'energy':E,'mu':mu,'mu_en':mu_en})
		self._mc_interp, self._mc_en_interp = None, None


	def _set_weights(self, elements, atom_weights=None, mass_weights=None):
		amu = pd.read_sql('SELECT * FROM weights', _get_connection('ziegler'))
		Zs = [ELEMENTS.index(el) for el in elements]

		if mass_weights is None:
			mass_weights = np.array([amu[amu['Z']==z]['amu'][z-1]*atom_weights[n] for n,z in enumerate(Zs)])
		elif atom_weights is None:
			atom_weights = np.array([mass_weights[n]/amu[amu['Z']==z]['amu'][z-1] for n,z in enumerate(Zs)])

		atom_weights, mass_weights = atom_weights/np.sum(atom_weights), mass_weights/np.sum(mass_weights)
		self.weights = pd.DataFrame({'element':elements, 'Z':Zs, 'atom_weight':atom_weights, 'mass_weight':mass_weights}, 
								columns=['element','Z','atom_weight','mass_weight'])


	def __str__(self):
		return self.name

	def saveas(self, filename, replace=False):
		"""Save the compound definition to a file

		The weights and density of the compound can be saved to one of
		the following file formats: .csv, .json, .db.  If the file exists,
		the data will be appended, unless `replace=True`, in which case
		the file will be replaced.  If a definition for the compound exists
		in the file already, it will be replaced.

		Parameters
		----------
		filename : str
			Filename where the compound will be saved.  Available formats
			are .csv, .json and .db.

		replace : bool, optional
			If `True`, replace the file if it exists.  Default `False`, which
			appends the data to the file.

		Examples
		--------
		>>> cm = ci.Compound('Brass', weights={'Zn':-33,'Cu':-66})
		>>> cm.saveas('brass.csv')
		>>> cm = ci.Compound('Water', weights={'H':2, 'O':1}, density=1.0)
		>>> cm.saveas('water.json')

		"""
		wts = self.weights.copy()
		wts['compound'] = self.name
		if self.density is not None:
			wts['density'] = self.density

		if filename.endswith('.csv'):
			if os.path.exists(filename) and not replace:
				df = pd.read_csv(filename, header=0)
				df = df[df['compound']!=self.name]
				df = pd.concat([df, wts])
				df.to_csv(filename, index=False)
			else:
				wts.to_csv(filename, index=False)

		if filename.endswith('.db'):
			if os.path.exists(filename) and not replace:
				con = _get_connection(filename)
				df = pd.read_sql('SELECT * FROM weights', con)
				df = df[df['compound']!=self.name]
				df = pd.concat([df, wts])
				df.to_sql('weights', con, if_exists='replace', index=False)
			else:
				wts.to_sql('weights', _get_connection(filename), if_exists='replace', index=False)

		if filename.endswith('.json'):
			if os.path.exists(filename) and not replace:
				df = pd.read_json(filename, orient='records')
				df = df[df['compound']!=self.name][wts.columns]
				df = pd.concat([df, wts])
				json.dump(json.loads(df.to_json(orient='records')), open(filename, 'w'), indent=4)
			else:
				json.dump(json.loads(wts.to_json(orient='records')), open(filename, 'w'), indent=4)

	def mu(self, energy):
		"""Mass-attenuation coefficient

		Interpolates the mass-attenuation coefficient, mu/rho,
		for the compound along the input energy grid.

		Parameters
		----------
		energy : array_like
			The incident photon energy, in keV.

		Returns
		-------
		mu : np.ndarray
			Mass attenuation coefficient, mu/rho, in cm^2/g.

		Examples
		--------
		>>> cm = ci.Compound('H2O')
		>>> print(cm.mu(200))
		0.13703928393005832

		"""

		if self._mc_interp is None:
			self._mc_interp = interp1d(np.log(self.mass_coeff['energy']), np.log(self.mass_coeff['mu']), bounds_error=False, fill_value='extrapolate')
		return np.exp(self._mc_interp(np.log(energy)))

	def mu_en(self, energy):
		"""Mass energy-absorption coefficient

		Interpolates the mass-energy absorption coefficient, mu_en/rho,
		for the compound along the input energy grid.

		Parameters
		----------
		energy : array_like
			The incident photon energy, in keV.

		Returns
		-------
		mu_en : np.ndarray
			Mass energy absorption coefficient, mu_en/rho, in cm^2/g.

		Examples
		--------
		>>> cm = ci.Compound('H2O')
		>>> print(cm.mu_en(200))
		0.029671598667776862

		"""

		if self._mc_en_interp is None:
			self._mc_en_interp = interp1d(np.log(self.mass_coeff['energy']), np.log(self.mass_coeff['mu_en']), bounds_error=False, fill_value='extrapolate')
		return np.exp(self._mc_en_interp(np.log(energy)))

	def attenuation(self, energy, x, density=None):
		"""Photon attenuation in matter

		Calculate the attenuation factor I(x)/I_0 = e^(-mu*x) for a given
		photon energy (in keV) and slab thickness (in cm).

		Parameters
		----------
		energy : array_like
			Incident photon energy in keV.

		x : float
			Thickness of slab of given compound, in cm. 

		density : float, optional
			Density of the compound in g/cm^3.  Default behavior is to
			use `Compound.density`, which must be supplied at construction.

		Returns
		-------
		attenuation : numpy.ndarray
			The slab attenuation factor as an absolute number (i.e. from 0 to 1).
			E.g. if the incident intensity is I_0, the transmitted intensity I(x) 
			is I_0 times the attenuation factor.

		Examples
		--------
		>>> cm = ci.Compound('SS_316') # preset compound for 316 Stainless
		>>> print(cm.attenuation(511, x=0.3))
		0.8199829388434694
		>>> print(cm.attenuation(300, x=1.0, density=5.0))
		0.5752140388004373

		"""

		energy = np.asarray(energy, dtype=np.float64)
		x = np.asarray(x, dtype=np.float64)
		if density is None:
			density = self.density

		return np.exp(-self.mu(energy)*x*density)
		
	def S(self, energy, particle='p', density=None):
		"""Charged particle stopping power in matter

		Calculate the stopping power, S=-dE/dx, for a given ion as a 
		function of the ion energy in MeV.  Units of S are MeV/cm.  To return
		stopping power in units of MeV/(mg/cm^2), use option `density=1E-3`.
		The stopping power is calculated using the Element.S() methods for
		each element in cm.elements, added using Bragg's rule.

		Parameters
		----------
		energy : array_like
			Incident ion energy in MeV.

		particle : str, optional
			Incident ion.  For light ions, options are 'p' (default), 'd', 't', 'a' for proton, 
			deuteron, triton and alpha, respectively.  Additionally, heavy ions can be
			specified either by element or isotope, e.g. 'Fe', '40CA', 'U', 'Bi-209'. For
			light ions, the charge state is assumed to be fully stripped. For heavy ions
			the charge state is handled by a Bohr/Northcliffe parameterization consistent
			with the Anderson-Ziegler formalism.

		density : float, optional
			Density of the compound in g/cm^3.  Default behavior is to use
			`Compound.density`.  To return stopping power in units of MeV/(mg/cm^2), i.e.
			the mass-stopping power, use `density=1E-3`.

		Returns
		-------
		stopping_power : numpy.ndarray
			Stopping power, S=-dE/dx, for a given ion as a function of the 
			ion energy in MeV.  Units of S are MeV/cm.

		Examples
		--------
		>>> cm = ci.Compound('SrCO3', density=3.5)
		>>> print(cm.S(60.0))
		27.196387031247834
		>>> print(cm.S(55.0, density=1E-3)) ### S in MeV/(mg/cm^2)
		0.008307827781861116

		"""

		energy = np.asarray(energy, dtype=np.float64)
		if density is None:
			density = self.density

		return np.average([el.S(energy, particle=particle, density=1E-3) for el in self.elements], weights=self.weights['mass_weight'], axis=0)*1E3*density
		
	def range(self, energy, particle='p', density=None):
		"""Charged particle range in matter

		Calculates the charged particle range in the compound, in cm.  Incident
		energy should be in MeV, and the particle type definition is identical
		to `Compound.S()`.

		Parameters
		----------
		energy : array_like
			Incident ion energy in MeV.

		particle : str, optional
			Incident ion.  For light ions, options are 'p' (default), 'd', 't', 'a' for proton, 
			deuteron, triton and alpha, respectively.  Additionally, heavy ions can be
			specified either by element or isotope, e.g. 'Fe', '40CA', 'U', 'Bi-209'. For
			light ions, the charge state is assumed to be fully stripped. For heavy ions
			the charge state is handled by a Bohr/Northcliffe parameterization consistent
			with the Anderson-Ziegler formalism.

		density : float, optional
			Density of the compound in g/cm^3.  Default behavior is to use
			`Compound.density`, which must be supplied at construction.
		
		Returns
		-------
		range : np.ndarray
			Charged particle range in the compound, in cm.

		Examples
		--------
		>>> cm = ci.Compound('Fe') # same behavior as element
		>>> print(cm.range(60.0))
		0.5858151125192633
		>>> cm = ci.Compound('SS_316') # preset compound
		>>> print(cm.range(60.0))
		0.5799450918147814

		"""

		energy = np.asarray(energy, dtype=np.float64)
		
		dE = np.max(energy)/1E3
		E_min = min((np.min(energy), 1.0))
		E_grid = np.arange(E_min, np.max(energy)+dE, dE)

		S = self.S(E_grid, particle=particle, density=density)
		x = np.cumsum((1.0/S)*dE)
		return interp1d(np.log(E_grid), x, bounds_error=None, fill_value='extrapolate')(np.log(energy))
		
	def plot_mass_coeff(self, energy=None, **kwargs):
		"""Plot the mass-attenuation coefficient in the compound

		Creates a plot of the mass-attenuation coefficient (in cm^2/g)
		as a function of photon energy in keV.

		Parameters
		----------
		energy : array_like, optional
			Energy grid on which to plot, replacing the default energy grid.
			Units are in keV.
		
		Other Parameters
		----------------
		**kwargs
			Optional keyword arguments for plotting.  See the 
			plotting section of the curie API for a complete
			list of kwargs.

		Examples
		--------
		>>> cm = ci.Compound('Fe')
		>>> cm.plot_mass_coeff()
		>>> cm = ci.Compound('H2O')
		>>> cm.plot_mass_coeff(style='poster')

		"""

		if energy is None:
			energy, mu = self.mass_coeff['energy'], self.mass_coeff['mu']
			
		else:
			energy = np.asarray(energy, dtype=np.float64)
			mu = self.mu(energy)

		f,ax = _init_plot(**kwargs)

		ax.plot(energy, mu, label=r'$\mu/\rho$'+' ({})'.format(self.name))
		ax.set_xlabel('Photon Energy (keV)')
		ax.set_ylabel(r'Attenuation Coeff. (cm$^2$/g)')
		ax.set_xscale('log')
		ax.set_yscale('log')
		ax.legend()
		
		return _draw_plot(f, ax, **kwargs)
		
	def plot_mass_coeff_en(self, energy=None, **kwargs):
		"""Plot the mass energy-absorption coefficient in the compound

		Creates a plot of the mass energy-absorption coefficient (in cm^2/g)
		as a function of photon energy in keV.

		Parameters
		----------
		energy : array_like, optional
			Energy grid on which to plot, replacing the default energy grid.
			Units are in keV.

		Other Parameters
		----------------
		**kwargs
			Optional keyword arguments for plotting.  See the 
			plotting section of the curie API for a complete
			list of kwargs.

		Examples
		--------
		>>> cm = ci.Compound('Silicone') # preset compound
		
		Example plotting the mass-attenuation coefficient together with the mass
		energy-absorption coefficient, on the same axes.

		>>> f,ax = cm.plot_mass_coeff(return_plot=True)
		>>> cm.plot_mass_coeff_en(f=f, ax=ax)

		"""

		if energy is None:
			energy, mu = self.mass_coeff['energy'], self.mass_coeff['mu_en']
			
		else:
			energy = np.asarray(energy, dtype=np.float64)
			mu = self.mu_en(energy)

		f,ax = _init_plot(**kwargs)

		ax.plot(energy, mu, label=r'$\mu_{en}/\rho$'+' ({})'.format(self.name))
		ax.set_xlabel('Photon Energy (keV)')
		ax.set_ylabel(r'Attenuation Coeff. (cm$^2$/g)')
		ax.set_xscale('log')
		ax.set_yscale('log')
		ax.legend()

		return _draw_plot(f, ax, **kwargs)
		
		
	def plot_S(self, particle='p', energy=None, **kwargs):
		"""Plot the stopping power in the compound

		Creates a plot of the charged particle stopping power (in MeV/(mg/cm^2))
		in the compound as a function of the incident ion energy (in MeV).

		Parameters
		----------
		particle : str
			Incident ion.  For light ions, options are 'p' (default), 'd', 't', 'a' for proton, 
			deuteron, triton and alpha, respectively.  Additionally, heavy ions can be
			specified either by element or isotope, e.g. 'Fe', '40CA', 'U', 'Bi-209'.

		energy : array_like, optional
			Energy grid on which to plot, replacing the default energy grid.
			Units are in MeV.

		Other Parameters
		----------------
		**kwargs
			Optional keyword arguments for plotting.  See the 
			plotting section of the curie API for a complete
			list of kwargs.

		Examples
		--------
		>>> cm = ci.Compound('He') # same as element
		>>> cm.plot_S(particle='a')
		>>> cm = ci.Compound('Kapton')
		>>> cm.plot_S(particle='d')

		"""

		if energy is None:
			energy = 10.0**np.arange(-1.5, 2.8, 0.05)

		f,ax = _init_plot(**kwargs)
		ax.plot(energy, self.S(energy, particle=particle, density=1E-3), label=r'$-\frac{dE}{dx}$ ('+self.name+')')

		ax.set_xlabel('Incident Energy (MeV)')
		ax.set_ylabel(r'Stopping Power (MeV/(mg/cm$^2$))')
		ax.set_xscale('log')
		ax.legend()

		return _draw_plot(f, ax, **kwargs)
		
	def plot_range(self, particle='p', energy=None, density=None, **kwargs):
		"""Plot the charged particle range in the compound

		Creates a plot of the charged particle range (in cm)
		in the compound as a function of the incident ion energy (in MeV).

		Parameters
		----------
		particle : str
			Incident ion.  For light ions, options are 'p' (default), 'd', 't', 'a' for proton, 
			deuteron, triton and alpha, respectively.  Additionally, heavy ions can be
			specified either by element or isotope, e.g. 'Fe', '40CA', 'U', 'Bi-209'.

		energy : array_like, optional
			Energy grid on which to plot, replacing the default energy grid.
			Units are in MeV.

		density : float, optional
			Density of the compound in g/cm^3.  Default behavior is to use
			`Compound.density`.

		Other Parameters
		----------------
		**kwargs
			Optional keyword arguments for plotting.  See the 
			plotting section of the curie API for a complete
			list of kwargs.

		Examples
		--------
		>>> cm = ci.Compound('Bronze', weights={'Cu':-80, 'Sn':-20}, density=8.9)
		>>> f,ax = cm.plot_range(return_plot=True)
		>>> cm.plot_range(particle='d', f=f, ax=ax)

		"""

		if energy is None:
			energy = 10.0**np.arange(-1.5, 2.8, 0.05)

		f,ax = _init_plot(**kwargs)
		ax.plot(energy, self.range(energy, particle=particle, density=density), label='Range ({})'.format(self.name))

		ax.set_xlabel('Incident Energy (MeV)')
		ax.set_ylabel('Range (cm)')
		ax.set_xscale('log')
		ax.set_yscale('log')
		ax.legend()

		return _draw_plot(f, ax, **kwargs)


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
try:
	from setuptools import setup, find_packages
	from setuptools.command.install import install as _install
except:
	from distutils.core import setup
	from distutils.command.install import install as _install



def _post_install(loc):
	path = lambda db: os.path.realpath(os.path.join(loc,'curie','data',db))

	if not os.path.isdir(path('')):
		os.mkdir(path(''))

	for fl in ['wwd6b1gk2ge5tgt/decay.db','tkndjqs036piojm/endf.db','zkoi6t2jicc9yqs/tendl.db','x2vfjr7uv7ffex5/tendl_d_rp.db','n0jjc0dv61j9of9/tendl_n_rp.db',
				'ib2a5lrhiwkcro5/tendl_p_rp.db','6jy0lns3vfun5vi/ziegler.db','lzn8zs6y8zu3v0s/iaea_monitors.db','34sgcvt8n57b0aw/IRDFF.db']:
		fnm = fl.split('/')[1]
		if not os.path.isfile(path(fnm)):
			try:
				import urllib2
			except:
				import urllib.request as urllib2
			try:
				print('Downloading {}'.format(fnm))
				with open(path(fnm),'wb') as f:
					f.write(urllib2.urlopen('https://www.dropbox.com/s/{}?dl=1'.format(fl)).read())
			except:
				print('ERROR: Unable to download {}. See https://jtmorrell.github.io/curie/build/html/quickinstall.html for more help.'.format(fnm))


class install(_install):
	def run(self):
		_install.run(self)
		self.execute(_post_install, (self.install_lib,), msg="Downloading nuclear data files...")

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='curie',
	  version='0.0.32',
	  description='Curie is a python toolkit to aid in the analysis of experimental nuclear data.',
	  long_description=long_description,
	  long_description_content_type="text/markdown",
	  url='https://github.com/jtmorrell/curie',
	  author='Jonathan Morrell',
	  author_email='jmorrell@berkeley.edu',
	  license='MIT',
	  packages=find_packages(),
	  include_package_data=True,
	  cmdclass={'install': install})#, 
	  #install_requires=['numpy', 'matplotlib', 'scipy', 'pandas'])